import importlib.util
from importlib.metadata import entry_points, EntryPoint
from os.path import join, dirname
import sys
from typing import Dict, List, Optional, Set
import warnings


def _hacky_entrypoint_loader(module_name):
    # load the root parent spec
    pkgs = module_name.split(".")
    spec = importlib.util.find_spec(pkgs[0])
    assert spec is not None and spec.origin is not None

    # even namespace packages are installed in the path, so just guess
    # ... and maybe it works?
    fname = join(dirname(spec.origin), *pkgs[1:]) + ".py"
    spec = importlib.util.spec_from_file_location(module_name, fname)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


class PkgCfg:
    """
    Contains information about an installed package that uses robotpy-build
    """

    def __init__(self, entry_point: EntryPoint):
        try:
            self.module = entry_point.load()
        except Exception as e:
            try:
                mod = (
                    entry_point.module
                    if hasattr(entry_point, "module")
                    else entry_point.value
                )
                self.module = _hacky_entrypoint_loader(mod)
            except Exception:
                raise e

        self.name = entry_point.name

        # could deduce this, but this is probably fine
        self.libinit_import = getattr(self.module, "libinit_import", None)
        self.depends = getattr(self.module, "depends", [])
        self.pypi_package = getattr(self.module, "pypi_package", None)
        self.package_name = getattr(self.module, "package_name", None)
        self.static_lib = getattr(self.module, "static_lib", False)

    def get_include_dirs(self) -> Optional[List[str]]:
        """
        Include directories provided by this module
        """
        fn = getattr(self.module, "get_include_dirs", None)
        if fn:
            return fn()
        return None

    def get_library_dirs(self) -> Optional[List[str]]:
        """
        Directories where libraries reside
        """
        fn = getattr(self.module, "get_library_dirs", None)
        if fn:
            return fn()
        return None

    def get_library_dirs_rel(self) -> Optional[List[str]]:
        """
        Directories where libraries reside, relative to package
        """
        fn = getattr(self.module, "get_library_dirs_rel", None)
        if fn:
            return fn()
        return None

    def get_library_names(self) -> Optional[List[str]]:
        """
        Names of libraries provided (for linking)
        """
        fn = getattr(self.module, "get_library_names", None)
        if fn:
            return fn()
        return None

    def get_extra_objects(self) -> Optional[List[str]]:
        """
        Names of extra objects to link in
        """
        fn = getattr(self.module, "get_extra_objects", None)
        if fn:
            return fn()
        return None

    def get_library_full_names(self) -> Optional[List[str]]:
        """
        Full names of libraries provided (needed for OSX support)
        """
        fn = getattr(self.module, "get_library_full_names", None)
        if fn:
            return fn()
        return None

    def get_type_casters(self, casters: Dict[str, str]) -> None:
        """
        Legacy type caster information
        """
        t = {}
        r = self.get_type_casters_cfg(t)
        for k, v in t.items():
            if "hdr" in v:
                casters[k] = v["hdr"]
        return r

    def get_type_casters_cfg(self, casters: Dict[str, str]) -> None:
        """
        Type caster headers provided

        key: type name
        value: a dict with keys:
            hdr: header file
            darg: force default arg

        """
        fn = getattr(self.module, "get_type_casters_cfg", None)
        if fn:
            return fn(casters)
        fn = getattr(self.module, "get_type_casters", None)
        if fn:
            t = {}
            r = fn(t)
            casters.update({k: {"hdr": v} for k, v in t.items()})
            return r


class PkgCfgProvider:
    """
    Retrieves information about robotpy-build packages

    Warning: Not to be confused with 'pkg-config'
    """

    def __init__(self):
        self.pkgs = {}

    def detect_pkgs(self) -> None:
        """
        Detect and load packages under the robotpybuild entry point group.
        Only loads packages that are dependencies.
        """
        deps_names = set().union(*[pkg.depends for pkg in self.pkgs.values()])
        ep_ret = entry_points()

        # Python 3.8/3.9
        if isinstance(ep_ret, dict):
            all_entry_points = ep_ret.get("robotpybuild", [])
        else:
            all_entry_points = [e for e in entry_points() if e.group == "robotpybuild"]

        # Only load the dependencies of the package we're building.
        # If we load the [package being built], then the current build will fail.
        # If we load a package that depends on the [package being built],
        # then the [package being built] will be loaded and the current build will fail.
        run_loop = True
        while run_loop:
            run_loop = False
            for ep in all_entry_points:
                if ep.name in self.pkgs:  # Prevents loading the package being built
                    continue
                if ep.name not in deps_names and ep.name != "robotpy-build":
                    continue
                try:
                    pkg = PkgCfg(ep)
                except Exception as e:
                    warnings.warn(f"Error loading entry point {ep.name}: {e}")
                else:
                    self.add_pkg(pkg)
                    deps_names |= set(pkg.depends)
                    run_loop = True

    def add_pkg(self, pkg: PkgCfg) -> None:
        self.pkgs[pkg.name] = pkg

    def get_pkg(self, name: str) -> PkgCfg:
        try:
            return self.pkgs[name]
        except KeyError:
            raise KeyError("robotpy-build package '%s' not installed" % name)

    def get_all_deps(self, name: str) -> Set[PkgCfg]:
        deps: Set[PkgCfg] = set()

        def _get(name: str):
            pkg = self.get_pkg(name)
            if pkg in deps:
                return pkg
            deps.add(pkg)
            for dep in pkg.depends:
                _get(dep)
            return pkg

        pkg = _get(name)
        deps.remove(pkg)
        return deps



#
# Uses cxxheaderparser to parse a header file and outputs a HeaderContext
# suitable for use with the autowrap templates
#

import dataclasses
import pathlib
import re
import sys
import typing
from keyword import iskeyword

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    Protocol = object  # pragma: no cover


import sphinxify
from cxxheaderparser.options import ParserOptions
from cxxheaderparser.parser import CxxParser
from cxxheaderparser.parserstate import (
    ClassBlockState,
    ExternBlockState,
    NamespaceBlockState,
    NonClassBlockState,
)
from cxxheaderparser.tokfmt import tokfmt
from cxxheaderparser.types import (
    AnonymousName,
    Array,
    ClassDecl,
    Concept,
    DecoratedType,
    DeductionGuide,
    EnumDecl,
    Field,
    ForwardDecl,
    FriendDecl,
    Function,
    FunctionType,
    FundamentalSpecifier,
    Method,
    MoveReference,
    NamespaceAlias,
    NameSpecifier,
    Parameter,
    Pointer,
    PQName,
    PQNameSegment,
    Reference,
    TemplateInst,
    Type,
    Typedef,
    UsingAlias,
    UsingDecl,
    Value,
    Variable,
)

from ..config.autowrap_yml import (
    AutowrapConfigYaml,
    BufferData,
    BufferType,
    ClassData,
    EnumValue,
    FunctionData,
    ParamData,
    PropAccess,
    ReturnValuePolicy,
)
from .generator_data import GeneratorData, OverloadTracker
from .context import (
    BaseClassData,
    ClassContext,
    ClassTemplateData,
    Documentation,
    EnumContext,
    EnumeratorContext,
    FunctionContext,
    GeneratedLambda,
    HeaderContext,
    ParamCategory,
    ParamContext,
    PropContext,
    TemplateInstanceContext,
    TrampolineData,
)


class HasSubpackage(Protocol):
    subpackage: typing.Optional[str]


class HasDoc(Protocol):
    doc: str
    doc_append: str


class HasNameData(Protocol):
    rename: str


# TODO: this isn't the best solution
def _gen_int_types():
    for i in ("int", "uint"):
        for j in ("", "_fast", "_least"):
            for k in ("8", "16", "32", "64"):
                yield f"{i}{j}{k}_t"
    yield "intmax_t"
    yield "uintmax_t"


_int32_types = frozenset(_gen_int_types())


_rvp_map = {
    ReturnValuePolicy.TAKE_OWNERSHIP: "py::return_value_policy::take_ownership",
    ReturnValuePolicy.COPY: "py::return_value_policy::copy",
    ReturnValuePolicy.MOVE: "py::return_value_policy::move",
    ReturnValuePolicy.REFERENCE: "py::return_value_policy::reference",
    ReturnValuePolicy.REFERENCE_INTERNAL: "py::return_value_policy::reference_internal",
    ReturnValuePolicy.AUTOMATIC: "",
    ReturnValuePolicy.AUTOMATIC_REFERENCE: "py::return_value_policy::automatic_reference",
}

# fmt: off
_operators = {
    # binary
    "-", "+", "*", "/", "%", "&", "^", "==", "!=", "|", ">", ">=", "<", "<=",
    # inplace
    "+=", "-=", "*=", "/=", "%=", "&=", "^=", "|=",
}
# fmt: on

_type_caster_seps = re.compile(r"[<>\(\)]")

_qualname_bad = ":<>="
_qualname_trans = str.maketrans(_qualname_bad, "_" * len(_qualname_bad))


_default_param_data = ParamData()
_default_enum_value = EnumValue()


def _is_fundamental(n: PQNameSegment):
    return isinstance(n, FundamentalSpecifier) or (
        isinstance(n, NameSpecifier) and n.name in _int32_types
    )


def _is_prop_readonly(dt: DecoratedType) -> bool:
    t: typing.Union[DecoratedType, FunctionType] = dt
    while True:
        if isinstance(t, Array):
            return False
        elif isinstance(t, (FunctionType, MoveReference, Reference)):
            return False
        elif isinstance(t, Type):
            return not _is_fundamental(t.typename.segments[-1])
        elif isinstance(t, Pointer):
            t = t.ptr_to
        else:
            assert False


def _count_and_unwrap(
    dt: DecoratedType,
) -> typing.Tuple[typing.Union[Array, FunctionType, Type], int, int, bool]:
    ptrs = 0
    refs = 0
    const = False
    t: typing.Union[DecoratedType, FunctionType] = dt
    while True:
        if isinstance(t, Type):
            const = const or t.const
            return t, ptrs, refs, const
        elif isinstance(t, (Array, FunctionType)):
            return t, ptrs, refs, const
        elif isinstance(t, Pointer):
            ptrs += 1
            const = const or t.const
            t = t.ptr_to
        elif isinstance(t, Reference):
            refs += 1
            t = t.ref_to
        elif isinstance(t, MoveReference):
            refs += 2
            t = t.moveref_to
        else:
            assert False


def _fmt_base_name(
    typename: PQName,
) -> typing.Tuple[str, str, str, str, str, typing.List[str]]:
    all_parts = []
    nameonly_parts = []

    for segment in typename.segments[:-1]:
        if not isinstance(segment, NameSpecifier):
            raise ValueError(f"base name formatting will fail for {typename.format()}")
        all_parts.append(segment.format())
        nameonly_parts.append(segment.name)

    last_segment = typename.segments[-1]
    if not isinstance(last_segment, NameSpecifier):
        raise ValueError(f"base name formatting will fail for {typename.format()}")

    nameonly_parts.append(last_segment.name)

    if last_segment.specialization:
        most_parts = all_parts[:-1]
        ns_parts = all_parts[:-1]
        all_parts.append(last_segment.format())
        most_parts.append(last_segment.name)
        tparam_list = [arg.format() for arg in last_segment.specialization.args]
    else:
        ns_parts = all_parts[:]
        all_parts.append(last_segment.name)
        most_parts = all_parts
        tparam_list = []

    return (
        last_segment.name,
        "::".join(most_parts),
        "::".join(all_parts),
        "::".join(nameonly_parts),
        "::".join(ns_parts),
        tparam_list,
    )

    # returns all names, with specialization for everything except the last
    # returns all names + specialization
    # returns list of all specialization parameters

    # cpp_name, cpp_name_w_templates, tparam_list

    assert False


T = typing.TypeVar("T")


def _fmt_nameonly(typename: PQName) -> str:
    """
    Not particularly general
    """
    parts = []
    for segment in typename.segments:
        if not isinstance(segment, NameSpecifier):
            return ""
        parts.append(segment.name)
    return "::".join(parts)


def _fmt_array_size(t: Array) -> typing.Optional[int]:
    if t.size:
        sz = tokfmt(t.size.tokens)
        try:
            return int(sz)
        except ValueError:
            pass

    return None


@dataclasses.dataclass
class _ReturnParamContext:
    #: was x_type
    cpp_type: str

    #: If this is an out parameter, the name of the parameter
    cpp_retname: str


#
# Visitor implementation
#


class ClassStateData(typing.NamedTuple):
    ctx: ClassContext
    cls_key: str
    data: ClassData

    typealias_names: typing.Set[str]

    # have to defer processing these
    defer_protected_methods: typing.List[Method]
    defer_private_nonvirtual_methods: typing.List[Method]
    defer_private_virtual_methods: typing.List[Method]
    defer_protected_fields: typing.List[Field]

    # Needed for trampoline
    template_argument_list: str
    base_template_params: str
    base_template_args: str

    # See j2_context::TrampolineData for these
    methods_to_disable: typing.List[FunctionContext]
    virtual_methods: typing.List[FunctionContext]
    protected_constructors: typing.List[FunctionContext]
    non_virtual_protected_methods: typing.List[FunctionContext]


Context = typing.Union[str, ClassStateData]

# define what user_data we store in each state type
# - classes are ClassStateData
# - everything else is the current namespace name
AWExternBlockState = ExternBlockState[str, str]
AWNamespaceBlockState = NamespaceBlockState[str, str]
AWClassBlockState = ClassBlockState[ClassStateData, Context]

AWState = typing.Union[AWExternBlockState, AWNamespaceBlockState, AWClassBlockState]
AWNonClassBlockState = typing.Union[AWExternBlockState, AWNamespaceBlockState]


class AutowrapVisitor:
    """
    Collects the results of parsing a header file
    """

    types: typing.Set[str]
    user_types: typing.Set[str]

    def __init__(
        self,
        hctx: HeaderContext,
        gendata: GeneratorData,
        casters: typing.Dict[str, typing.Dict[str, typing.Any]],
        report_only: bool,
    ) -> None:
        self.hctx = hctx
        self.gendata = gendata
        self.user_cfg = gendata.data
        self.report_only = report_only
        self.casters = casters
        self.types = set()
        self.user_types = set()

    #
    # Visitor interface
    #

    def on_parse_start(self, state: AWNamespaceBlockState) -> None:
        state.user_data = ""

    def on_pragma(self, state: AWState, content: Value) -> None:
        pass

    def on_include(self, state: AWState, filename: str) -> None:
        pass

    def on_extern_block_start(self, state: AWExternBlockState) -> None:
        state.user_data = state.parent.user_data

    def on_extern_block_end(self, state: AWExternBlockState) -> None:
        pass

    def on_namespace_start(self, state: AWNamespaceBlockState) -> None:
        parent_ns = typing.cast(str, state.parent.user_data)
        names = state.namespace.names
        if not names:
            # anonymous namespace items are referenced using the parent
            ns = parent_ns
        elif not parent_ns:
            ns = "::".join(names)
        else:
            ns = f"{parent_ns}::{'::'.join(names)}"

        state.user_data = ns

        if ns not in self.hctx.namespaces:
            self.hctx.namespaces.append(ns)

    def on_namespace_end(self, state: AWNamespaceBlockState) -> None:
        pass

    def on_namespace_alias(
        self, state: AWNonClassBlockState, alias: NamespaceAlias
    ) -> None:
        # TODO: add to some sort of resolver?
        pass

    def on_concept(self, state: AWNonClassBlockState, concept: Concept) -> None:
        pass

    def on_forward_decl(self, state: AWState, fdecl: ForwardDecl) -> None:
        # TODO: add to some sort of resolver?
        pass

    def on_template_inst(self, state: AWState, inst: TemplateInst) -> None:
        pass

    def on_variable(self, state: AWState, v: Variable) -> None:
        # TODO: robotpy-build doesn't wrap global variables at this time
        pass

    def on_function(self, state: AWNonClassBlockState, fn: Function) -> None:
        # operators that aren't class members aren't rendered
        if fn.operator:
            return

        # ignore functions with complicated names
        fn_name = self._get_fn_name(fn)
        if not fn_name:
            return

        data, overload_tracker = self.gendata.get_function_data(fn_name, fn)
        if data.ignore:
            return

        scope_var = self._get_module_var(data)
        fctx = self._on_fn_or_method(
            fn, data, fn_name, scope_var, False, overload_tracker
        )
        fctx.namespace = state.user_data
        self.hctx.functions.append(fctx)

    def on_method_impl(self, state: AWNonClassBlockState, method: Method) -> None:
        # we only wrap methods when defined in a class
        pass

    def on_typedef(self, state: AWState, typedef: Typedef) -> None:
        pass

    def on_using_namespace(
        self, state: AWNonClassBlockState, namespace: typing.List[str]
    ) -> None:
        pass

    def on_using_alias(self, state: AWState, using: UsingAlias) -> None:
        self._add_type_caster(using.type)

        # autodetect embedded using directives, but don't override anything
        # the user specifies
        # - these are in block scope, so they cannot include templates
        if (
            isinstance(state, ClassBlockState)
            and using.access == "public"
            and using.template is None
            and using.alias not in state.user_data.typealias_names
        ):
            ctx = state.user_data.ctx
            ctx.auto_typealias.append(
                f"using {using.alias} [[maybe_unused]] = typename {ctx.full_cpp_name}::{using.alias}"
            )

    def on_using_declaration(self, state: AWState, using: UsingDecl) -> None:
        self._add_type_caster_pqname(using.typename)

        if using.access is None:
            self.hctx.using_declarations.append(using.typename)
        elif isinstance(state, ClassBlockState):
            # A using declaration might bring in a colliding name for a function,
            # so mark it as overloaded
            lseg = using.typename.segments[-1]
            if isinstance(lseg, NameSpecifier):
                cdata = state.user_data
                name = lseg.name
                self.gendata.add_using_decl(
                    name, cdata.cls_key, cdata.data, state.access == "private"
                )

    #
    # Enums
    #

    def on_enum(self, state: AWState, enum: EnumDecl) -> None:
        # If an enum name has more than one component, that's weird and we're
        # not going to support it for now. Who forward declares enums anyways?
        name_segs = enum.typename.segments
        assert len(name_segs) > 0

        if len(name_segs) > 1:
            return
        elif isinstance(name_segs[0], NameSpecifier):
            ename = name_segs[0].name
        elif isinstance(name_segs[0], AnonymousName):
            ename = ""
        else:
            # something else weird we can't support for now
            return

        user_data = state.user_data

        if isinstance(user_data, str):
            # global enum

            enum_data = self.gendata.get_enum_data(ename)
            if enum_data.ignore:
                return

            ctxlist = self.hctx.enums
            var_name = f"enum{len(ctxlist)}"
            enum_scope = user_data
            scope_var = self._get_module_var(enum_data)
        else:
            # per-class -- ignore private/protected enums
            if enum.access != "public":
                return

            enum_data = self.gendata.get_cls_enum_data(
                ename, user_data.cls_key, user_data.data
            )
            if enum_data.ignore:
                return

            cls_ctx = user_data.ctx

            if ename:
                ctxlist = cls_ctx.enums
                var_name = f"{cls_ctx.var_name}_enum{len(ctxlist)}"
            else:
                ctxlist = cls_ctx.unnamed_enums
                var_name = f"{cls_ctx.var_name}_enum_u{len(ctxlist)}"

            enum_scope = cls_ctx.full_cpp_name
            scope_var = cls_ctx.var_name

        value_prefix = None
        strip_prefixes = []
        values: typing.List[EnumeratorContext] = []

        py_name = ""
        full_cpp_name = ""

        if ename:
            full_cpp_name = f"{enum_scope}::{ename}"
            py_name = self._make_py_name(ename, enum_data)

            value_prefix = enum_data.value_prefix
            if not value_prefix:
                value_prefix = ename

            strip_prefixes = [f"{value_prefix}_", value_prefix]
        else:
            full_cpp_name = enum_scope

        for v in enum.values:
            name = v.name
            v_data = enum_data.values.get(name, _default_enum_value)
            if v_data.ignore:
                continue

            values.append(
                EnumeratorContext(
                    full_cpp_name=f"{full_cpp_name}::{name}",
                    py_name=self._make_py_name(name, v_data, strip_prefixes),
                    doc=self._process_doc(v.doxygen, v_data, append_prefix="  "),
                )
            )

        ctxlist.append(
            EnumContext(
                scope_var=scope_var,
                var_name=var_name,
                cpp_name=ename,
                full_cpp_name=full_cpp_name,
                py_name=py_name,
                values=values,
                doc=self._process_doc(enum.doxygen, enum_data),
                arithmetic=enum_data.arithmetic,
                inline_code=enum_data.inline_code,
            )
        )

    #
    # Class/union/struct
    #

    def on_class_start(self, state: AWClassBlockState) -> typing.Optional[bool]:
        # parse everything there is to parse about the declaration
        # of this class, and append it

        # ignore non-public nested classes or classes that have ignored parents
        if state.typedef or (
            isinstance(state.parent, ClassBlockState)
            and state.parent.access != "public"
        ):
            return False

        cls_name_result = self._process_class_name(state)
        if cls_name_result is None:
            return False

        cls_key, cls_name, cls_namespace, parent_ctx = cls_name_result
        class_data = self.gendata.get_class_data(cls_key)

        # Ignore explicitly ignored classes
        if class_data.ignore:
            return False

        for typename in class_data.force_type_casters:
            self._add_user_type_caster(typename)

        class_decl = state.class_decl
        var_name = f"cls_{cls_name}"

        if parent_ctx:
            cls_qualname = f"{parent_ctx.full_cpp_name}::{cls_name}"
            dep_cpp_name = f"{parent_ctx.dep_cpp_name}::{cls_name}"
            scope_var = parent_ctx.var_name
        else:
            cls_qualname = f"{cls_namespace}::{cls_name}"
            dep_cpp_name = cls_qualname
            scope_var = self._get_module_var(class_data)

        cls_cpp_identifier = cls_qualname.translate(_qualname_trans)

        #
        # Process inheritance
        #

        bases, pybase_params = self._process_class_bases(
            cls_namespace, cls_name, class_decl, class_data
        )

        # All names used here should not have any specializations in them
        self.hctx.class_hierarchy[dep_cpp_name] = [
            base.dep_cpp_name for base in bases
        ] + class_data.force_depends

        #
        # Process template parameters
        #

        # <N, .. >
        template_argument_list = ""
        # <typename N, .. >
        template_parameter_list = ""

        template_data: typing.Optional[ClassTemplateData] = None

        if class_data.template_params:
            if class_data.subpackage:
                raise ValueError(
                    f"{cls_name}: classes with subpackages must define subpackage on template instantiation"
                )

            template_args = []
            template_params = []

            base_template_args = []
            base_template_params = []

            # TODO: should be able to remove this parsing since cxxheaderparser
            #       can figure it out for us

            for param in class_data.template_params:
                if " " in param:
                    arg = param.split(" ", 1)[1]
                else:
                    arg = param
                    param = f"typename {param}"

                template_args.append(arg)
                template_params.append(param)

                if arg in pybase_params:
                    base_template_args.append(arg)
                    base_template_params.append(param)

            template_argument_list = ", ".join(template_args)
            template_parameter_list = ", ".join(template_params)

            template_data = ClassTemplateData(
                argument_list=template_argument_list,
                parameter_list=template_parameter_list,
                inline_code=class_data.template_inline_code,
            )

            cls_qualname = f"{cls_qualname}<{template_argument_list}>"

            base_template_params_s = ", ".join(base_template_params)
            base_template_args_s = ", ".join(base_template_args)
        else:
            base_template_params_s = ""
            base_template_args_s = ""

        if not self.report_only:
            if class_decl.template:
                if template_parameter_list == "":
                    raise ValueError(
                        f"{cls_name}: must specify template_params for templated class, or ignore it"
                    )
            else:
                if template_parameter_list != "":
                    raise ValueError(
                        f"{cls_name}: cannot specify template_params for non-template class"
                    )

        #
        # Other stuff
        #

        if class_data.is_polymorphic is not None:
            is_polymorphic = class_data.is_polymorphic
        else:
            # bad assumption? probably
            is_polymorphic = len(class_decl.bases) > 0

        doc = self._process_doc(class_decl.doxygen, class_data)
        py_name = self._make_py_name(cls_name, class_data)

        constants: typing.List[typing.Tuple[str, str]] = []
        for constant in class_data.constants:
            name = constant.split("::")[-1]
            constants.append((name, constant))

        # do logic for extracting user defined typealiases here
        # - these are at class scope, so they can include template
        typealias_names: typing.Set[str] = set()
        user_typealias: typing.List[str] = []
        self._extract_typealias(class_data.typealias, user_typealias, typealias_names)

        ctx = ClassContext(
            parent=parent_ctx,
            namespace=cls_namespace,
            cpp_name=cls_name,
            full_cpp_name=cls_qualname,
            full_cpp_name_identifier=cls_cpp_identifier,
            dep_cpp_name=dep_cpp_name,
            py_name=py_name,
            scope_var=scope_var,
            var_name=var_name,
            nodelete=class_data.nodelete,
            final=class_decl.final,
            doc=doc,
            bases=bases,
            template=template_data,
            user_typealias=user_typealias,
            constants=constants,
            inline_code=class_data.inline_code or "",
            force_multiple_inheritance=class_data.force_multiple_inheritance,
            is_polymorphic=is_polymorphic,
        )

        # Add to parent class or global class list
        if parent_ctx:
            parent_ctx.child_classes.append(ctx)
        else:
            self.hctx.classes.append(ctx)

        # Store for other events to use
        state.user_data = ClassStateData(
            ctx=ctx,
            cls_key=cls_key,
            data=class_data,
            typealias_names=typealias_names,
            # Method data
            defer_protected_methods=[],
            defer_private_nonvirtual_methods=[],
            defer_private_virtual_methods=[],
            defer_protected_fields=[],
            # Trampoline data
            template_argument_list=template_argument_list,
            base_template_args=base_template_args_s,
            base_template_params=base_template_params_s,
            methods_to_disable=[],
            virtual_methods=[],
            protected_constructors=[],
            non_virtual_protected_methods=[],
        )

    def _process_class_name(
        self, state: AWClassBlockState
    ) -> typing.Optional[typing.Tuple[str, str, str, typing.Optional[ClassContext]]]:
        class_decl = state.class_decl
        segments = class_decl.typename.segments
        assert len(segments) > 0

        segment_names: typing.List[str] = []
        for segment in segments:
            if not isinstance(segment, NameSpecifier):
                raise ValueError(
                    f"not sure how to handle '{class_decl.typename.format()}'"
                )
            # ignore specializations for now
            if segment.specialization is not None:
                return None
            segment_names.append(segment.name)

        cls_name = segment_names[-1]
        extra_segments = "::".join(segment_names[:-1])

        parent_ctx: typing.Optional[ClassContext] = None

        parent = state.parent

        if not isinstance(parent, ClassBlockState):
            # easy case -- namespace is the next user_data up
            cls_key = cls_name
            cls_namespace = typing.cast(str, parent.user_data)
            if extra_segments:
                cls_namespace = f"{cls_namespace}::{extra_segments}"
        else:
            # Use things the parent already computed
            cdata = typing.cast(ClassStateData, parent.user_data)
            # parent: AWClassBlockState = state.parent
            parent_ctx = cdata.ctx
            # the parent context already computed namespace, so use that
            if extra_segments:
                cls_key = f"{cdata.cls_key}::{extra_segments}::{cls_name}"
            else:
                cls_key = f"{cdata.cls_key}::{cls_name}"
            cls_namespace = parent_ctx.namespace

        return cls_key, cls_name, cls_namespace, parent_ctx

    def _process_class_bases(
        self,
        cls_namespace: str,
        cls_name: str,
        class_decl: ClassDecl,
        class_data: ClassData,
    ) -> typing.Tuple[
        typing.List[BaseClassData],
        typing.Set[str],
    ]:
        bases: typing.List[BaseClassData] = []
        pybase_params: typing.Set[str] = set()
        ignored_bases = {ib: True for ib in class_data.ignored_bases}

        for base in class_decl.bases:
            if base.access == "private":
                continue

            (
                cpp_name,
                full_cpp_name,
                cpp_name_w_templates,
                dep_cpp_name,
                base_ns,
                tparam_list,
            ) = _fmt_base_name(base.typename)
            if ignored_bases.pop(cpp_name_w_templates, None):
                continue

            # Sometimes, we can't guess all the information about the base, so the
            # user needs to specify it explicitly.
            user_bqual = class_data.base_qualnames.get(full_cpp_name)
            if user_bqual:
                cpp_name_w_templates = user_bqual
                # TODO: sometimes need to add this to pybase_params, but
                # that would require parsing this more. Seems sufficiently
                # obscure, going to omit it for now.
                tp = user_bqual.find("<")
                if tp == -1:
                    full_cpp_name = user_bqual
                    template_params = ""
                else:
                    full_cpp_name = user_bqual[:tp]
                    template_params = user_bqual[tp + 1 : -1]
                dep_cpp_name = full_cpp_name
                ns_idx = full_cpp_name.rfind("::")
                if ns_idx == -1:
                    base_ns = ""
                else:
                    base_ns = full_cpp_name[:ns_idx]
            else:
                # TODO: we don't handle nested child classes with templates here
                #       ... but that has to be rather obscure?

                for param in tparam_list:
                    pybase_params.add(param)

                template_params = ", ".join(tparam_list)

                # If no explicit namespace specified, we assume base classes
                # live in the same namespace as the class
                if len(base.typename.segments) == 1:
                    base_ns = cls_namespace
                    full_cpp_name = f"{cls_namespace}::{full_cpp_name}"
                    cpp_name_w_templates = f"{cls_namespace}::{cpp_name_w_templates}"
                    dep_cpp_name = f"{cls_namespace}::{dep_cpp_name}"

            base_identifier = full_cpp_name.translate(_qualname_trans)

            if base_ns:
                base_ns = f"{base_ns}::"

            bases.append(
                BaseClassData(
                    cls_name=cpp_name,
                    namespace_=base_ns,
                    full_cpp_name=full_cpp_name,
                    full_cpp_name_w_templates=cpp_name_w_templates,
                    full_cpp_name_identifier=base_identifier,
                    dep_cpp_name=dep_cpp_name,
                    template_params=template_params,
                )
            )

        if not self.report_only and ignored_bases:
            bases_s = ", ".join(base.typename.format() for base in class_decl.bases)
            invalid_bases = ", ".join(ignored_bases.keys())
            raise ValueError(
                f"{cls_name}: ignored_bases contains non-existant bases "
                + f"{invalid_bases}; valid bases are {bases_s}"
            )

        return bases, pybase_params

    def on_class_field(self, state: AWClassBlockState, f: Field) -> None:
        # Ignore unnamed fields
        if not f.name:
            return

        access = f.access
        if access == "public":
            self._on_class_field(state, f, state.user_data.ctx.public_properties)
        elif access == "protected":
            state.user_data.defer_protected_fields.append(f)

    def _on_class_field(
        self, state: AWClassBlockState, f: Field, props: typing.List[PropContext]
    ) -> None:
        prop_name = f.name
        if prop_name is None:
            return

        propdata = self.gendata.get_cls_prop_data(
            prop_name, state.user_data.cls_key, state.user_data.data
        )
        if propdata.ignore:
            return

        self._add_type_caster(f.type)
        if propdata.rename:
            py_name = propdata.rename
        elif f.access != "public":
            py_name = f"_{prop_name}"
        elif iskeyword(prop_name):
            py_name = f"{prop_name}_"
        else:
            py_name = prop_name

        if propdata.access == PropAccess.AUTOMATIC:
            # const variables can't be written
            if f.constexpr or getattr(f.type, "const", False):
                prop_readonly = True
            # We assume that a struct intentionally has readwrite data
            # attributes regardless of type
            elif state.class_decl.classkey != "class":
                prop_readonly = False
            else:
                prop_readonly = _is_prop_readonly(f.type)
        else:
            prop_readonly = propdata.access == PropAccess.READONLY

        doc = self._process_doc(f.doxygen, propdata)

        array_size = None
        is_array = False

        if isinstance(f.type, Array):
            is_array = True
            array_size = _fmt_array_size(f.type)
            cpp_type = f.type.array_of.format()
        else:
            cpp_type = f.type.format()

        props.append(
            PropContext(
                py_name=py_name,
                cpp_name=prop_name,
                cpp_type=cpp_type,
                readonly=prop_readonly,
                doc=doc,
                array_size=array_size,
                array=is_array,
                reference=isinstance(f.type, Reference),
                static=f.static,
                bitfield=f.bits is not None,
            )
        )

        # If it's constexpr, insert into the binding scope
        if f.access == "public" and f.constexpr:
            cctx = state.user_data.ctx
            cctx.auto_typealias.append(
                f"static constexpr auto {prop_name} [[maybe_unused]] = {cctx.full_cpp_name}::{prop_name}"
            )

    def on_class_method(self, state: AWClassBlockState, method: Method) -> None:
        # This needs to only process enough about the method to answer things
        # that are needed in on_class_end. Some methods are only processed in
        # on_class_end if the answers are right
        cdata = state.user_data
        cctx = cdata.ctx

        if method.constructor:
            cctx.has_constructor = True
        is_polymorphic = method.virtual or method.override or method.final
        if is_polymorphic:
            cctx.is_polymorphic = True

        access = state.access
        if access == "public":
            # Go ahead and process public methods now
            self._on_class_method(state, method, cctx.wrapped_public_methods)
        elif access == "protected":
            cdata.defer_protected_methods.append(method)
        elif access == "private":
            if is_polymorphic:
                cdata.defer_private_virtual_methods.append(method)
            else:
                cdata.defer_private_nonvirtual_methods.append(method)

    def _on_class_method(
        self,
        state: AWClassBlockState,
        method: Method,
        methods: typing.List[FunctionContext],
    ) -> None:
        cdata = state.user_data
        cctx = cdata.ctx

        # I think this is always true?
        assert len(method.name.segments) == 1

        method_name = self._get_fn_name(method)
        if not method_name:
            return

        is_constructor = method.constructor
        is_override = method.override
        is_virtual = method.virtual or is_override

        operator = method.operator

        # Ignore some operators, deleted methods, destructors
        if (
            (operator and operator not in _operators)
            or method.destructor
            or method.deleted
        ):
            return

        # Also ignore move constructors and copy constructors
        if (
            is_constructor
            and len(method.parameters) == 1
            and self._is_copy_move_constructor(cctx, method.parameters[0].type)
        ):
            return

        is_final = method.final
        is_private = state.access == "private"

        method_data, overload_tracker = self.gendata.get_function_data(
            method_name,
            method,
            cdata.cls_key,
            cdata.data,
            state.access == "private",
        )
        if method_data.ignore:
            return

        fctx = self._on_fn_or_method(
            method,
            method_data,
            method_name,
            cdata.ctx.scope_var,
            state.access != "public",
            overload_tracker,
        )

        # Update class-specific method attributes
        fctx.is_constructor = is_constructor
        if is_constructor and method_data.rename and not method_data.cpp_code:
            fctx.cpp_return_type = cctx.full_cpp_name
            self._on_fn_make_lambda(method_data, fctx)
        if operator:
            fctx.operator = operator
            self.hctx.need_operators_h = True
            if method_data.no_release_gil is None:
                fctx.release_gil = False

            # Use cpp_code to setup the operator
            if fctx.cpp_code is None:
                if len(method.parameters) == 0:
                    fctx.cpp_code = f"{operator} py::self"
                else:
                    ptype, _, _, _ = _count_and_unwrap(method.parameters[0].type)
                    if (
                        isinstance(ptype, Type)
                        and isinstance(ptype.typename.segments[-1], NameSpecifier)
                        and ptype.typename.segments[-1].name == cdata.ctx.cpp_name
                    ):
                        # don't try to predict the type, use py::self instead
                        fctx.cpp_code = f"py::self {operator} py::self"
                    else:
                        fctx.cpp_code = f"py::self {operator} {fctx.all_params[0].cpp_type_no_const}()"

        if method.const:
            fctx.const = True
        if method.static:
            fctx.is_static_method = True
        if method.pure_virtual:
            fctx.is_pure_virtual = True
        if method.ref_qualifier:
            fctx.ref_qualifiers = method.ref_qualifier

        # automatically retain references passed to constructors if the
        # user didn't specify their own keepalive
        if is_constructor and method_data.keepalive is None:
            for i, pctx in enumerate(fctx.filtered_params):
                if pctx.full_cpp_type.endswith("&"):
                    fctx.keepalives.append((1, i + 2))

        # Update method lists
        if is_private and is_override:
            cdata.methods_to_disable.append(fctx)
        else:
            if is_final:
                cdata.methods_to_disable.append(fctx)

            # disable virtual method generation for functions with buffer
            # parameters (doing it correctly is hard, so we skip it)
            if is_virtual and not fctx.has_buffers:
                cdata.virtual_methods.append(fctx)

            if not is_private:
                if not fctx.ignore_py:
                    methods.append(fctx)

                if state.access == "protected":
                    if is_constructor:
                        cdata.protected_constructors.append(fctx)
                    elif not is_virtual:
                        cdata.non_virtual_protected_methods.append(fctx)

        # If the method has cpp_code defined, it must either match the function
        # signature of the method, or virtual_xform must be defined with an
        # appropriate conversion. If neither of these are true, it will lead
        # to difficult to diagnose errors at runtime. We add a static assert
        # to try and catch these errors at compile time
        need_vcheck = (
            is_virtual
            and method_data.cpp_code
            and not method_data.virtual_xform
            and not method_data.trampoline_cpp_code
            and not state.class_decl.final
            and not cdata.data.force_no_trampoline
        )
        if need_vcheck:
            cctx.vcheck_fns.append(fctx)
            self.hctx.has_vcheck = True

        # Check for user data errors
        if not self.report_only:
            if method_data.ignore_pure and not method.pure_virtual:
                raise ValueError(
                    f"{cdata.cls_key}::{method_name}: cannot specify ignore_pure for function that isn't pure"
                )

            if method_data.trampoline_cpp_code and not is_virtual:
                raise ValueError(
                    f"{cdata.cls_key}::{method_name}: cannot specify trampoline_cpp_code for a non-virtual method"
                )

            if method_data.virtual_xform and not is_virtual:
                raise ValueError(
                    f"{cdata.cls_key}::{method_name}: cannot specify virtual_xform for a non-virtual method"
                )

            # pybind11 doesn't support this, user must fix it
            if (
                method.ref_qualifier == "&&"
                and not method_data.ignore_py
                and not method_data.cpp_code
            ):
                raise ValueError(
                    f"{cdata.cls_key}::{method_name}: has && ref-qualifier which cannot be directly bound by pybind11, must specify cpp_code or ignore_py"
                )

    def _on_class_method_process_overload_only(
        self, state: AWClassBlockState, method: Method
    ):
        cdata = state.user_data

        method_name = self._get_fn_name(method)
        if not method_name:
            return

        self.gendata.get_function_data(
            method_name,
            method,
            cdata.cls_key,
            cdata.data,
            True,
        )

    def _is_copy_move_constructor(
        self, cctx: ClassContext, first_type_param: DecoratedType
    ) -> bool:
        if isinstance(first_type_param, Reference):
            t = first_type_param.ref_to
        elif isinstance(first_type_param, MoveReference):
            t = first_type_param.moveref_to
        else:
            return False

        if not isinstance(t, Type):
            return False

        last_seg = t.typename.segments[-1]
        if not isinstance(last_seg, NameSpecifier):
            return False

        if len(t.typename.segments) == 1:
            return last_seg.name == cctx.cpp_name
        else:
            # This isn't quite right, but probably rarely happens?
            param_name = _fmt_nameonly(t.typename)
            return param_name == cctx.full_cpp_name

    def on_class_friend(self, state: AWClassBlockState, friend: FriendDecl) -> None:
        pass

    def on_class_end(self, state: AWClassBlockState) -> None:
        # post-process the class data
        cdata = state.user_data
        ctx = cdata.ctx
        class_data = cdata.data

        # If there isn't already a constructor, add a default constructor
        # - was going to add a FunctionContext for it, but.. this is way easier
        ctx.add_default_constructor = (
            not ctx.has_constructor
            and not class_data.nodelete
            and not class_data.force_no_default_constructor
        )

        has_trampoline = (
            ctx.is_polymorphic
            and not state.class_decl.final
            and not class_data.force_no_trampoline
        )

        # process methods and fields
        if has_trampoline:
            state.access = "protected"
            methods = ctx.wrapped_protected_methods
            for m in cdata.defer_protected_methods:
                self._on_class_method(state, m, methods)

            props = ctx.protected_properties
            for f in cdata.defer_protected_fields:
                self._on_class_field(state, f, props)

            state.access = "private"
            unused = []
            for m in cdata.defer_private_virtual_methods:
                self._on_class_method(state, m, unused)

            self.hctx.classes_with_trampolines.append(ctx)

            tmpl = ""
            if cdata.template_argument_list:
                tmpl = f", {cdata.template_argument_list}"

            trampoline_cfg = f"{ctx.namespace}::PyTrampolineCfg_{ctx.cpp_name}<{cdata.template_argument_list}>"
            tname = f"{ctx.namespace}::PyTrampoline_{ctx.cpp_name}<typename {ctx.full_cpp_name}{tmpl}, typename {trampoline_cfg}>"
            tvar = f"{ctx.cpp_name}_Trampoline"

            ctx.trampoline = TrampolineData(
                full_cpp_name=tname,
                var=tvar,
                inline_code=class_data.trampoline_inline_code,
                tmpl_args=cdata.base_template_args,
                tmpl_params=cdata.base_template_params,
                methods_to_disable=cdata.methods_to_disable,
                virtual_methods=cdata.virtual_methods,
                protected_constructors=cdata.protected_constructors,
                non_virtual_protected_methods=cdata.non_virtual_protected_methods,
            )

        elif class_data.trampoline_inline_code is not None:
            raise ValueError(
                f"{cdata.cls_key} has trampoline_inline_code specified, but there is no trampoline!"
            )

        else:
            # still need to do minimal processing to add deferred functions
            # to the overload tracker, otherwise we won't handle it correctly
            for m in cdata.defer_protected_methods:
                self._on_class_method_process_overload_only(state, m)
            for m in cdata.defer_private_virtual_methods:
                self._on_class_method_process_overload_only(state, m)

        for m in cdata.defer_private_nonvirtual_methods:
            self._on_class_method_process_overload_only(state, m)

    def on_deduction_guide(
        self, state: NonClassBlockState, guide: DeductionGuide
    ) -> None:
        pass

    #
    # Function/method processing
    #

    def _on_fn_or_method(
        self,
        fn: Function,
        data: FunctionData,
        fn_name: str,
        scope_var: str,
        internal: bool,
        overload_tracker: OverloadTracker,
    ) -> FunctionContext:
        # if cpp_code is specified, don't release the gil unless the user
        # specifically asks for it
        if data.no_release_gil is None:
            release_gil = data.cpp_code is None
        else:
            release_gil = not data.no_release_gil

        all_params: typing.List[ParamContext] = []
        filtered_params: typing.List[ParamContext] = []
        keepalives = []

        has_out_param = False

        # Use this if one of the parameter types don't quite match
        param_override = data.param_override
        fn_disable_none = data.disable_none

        # keep track of param name changes so we can automatically update
        # documentation
        param_remap: typing.Dict[str, str] = {}

        #
        # Process parameters
        #

        for i, p in enumerate(fn.parameters):
            p_name = p.name
            if not p_name:
                p_name = f"param{i}"

            po = param_override.get(p_name, _default_param_data)

            pctx = self._on_fn_param(
                p,
                p_name,
                fn_disable_none,
                po,
                param_remap,
            )

            all_params.append(pctx)
            if not po.ignore:
                filtered_params.append(pctx)

            if pctx.category == ParamCategory.OUT:
                has_out_param = True

        return_value_policy = _rvp_map[data.return_value_policy]

        # Set up the function's name
        if data.rename:
            # user preference wins, of course
            py_name = data.rename
        elif isinstance(fn, Method) and fn.constructor:
            py_name = "__init__"
        else:
            # Python exposed function name converted to camelcase
            py_name = self._make_py_name(
                fn_name, data, is_operator=fn.operator is not None
            )
            if not py_name[:2].isupper():
                py_name = f"{py_name[0].lower()}{py_name[1:]}"

            if data.internal or internal:
                py_name = f"_{py_name}"

        doc = self._process_doc(fn.doxygen, data, param_remap=param_remap)

        # Allow the user to override our auto-detected keepalives
        if data.keepalive is not None:
            keepalives = data.keepalive

        # Check for user errors
        if not self.report_only:
            if fn.template:
                if data.template_impls is None and not data.cpp_code:
                    raise ValueError(
                        f"{fn_name}: must specify template impls for function template"
                    )
            else:
                if data.template_impls is not None:
                    raise ValueError(
                        f"{fn_name}: cannot specify template_impls for non-template functions"
                    )

        #
        # fn_retval is needed for gensig, vcheck assertions
        # - gensig is not computable here
        #
        fn_retval: typing.Optional[str] = None
        if fn.return_type:
            fn_retval = fn.return_type.format()
            self._add_type_caster(fn.return_type)

        fctx = FunctionContext(
            cpp_name=fn_name,
            doc=doc,
            scope_var=scope_var,
            # transforms
            py_name=py_name,
            cpp_return_type=fn_retval,
            all_params=all_params,
            filtered_params=filtered_params,
            has_buffers=bool(data.buffers),
            keepalives=keepalives,
            return_value_policy=return_value_policy,
            # info
            # vararg=fn.vararg,
            # user settings
            ignore_pure=data.ignore_pure,
            ignore_py=data.ignore_py,
            cpp_code=data.cpp_code,
            trampoline_cpp_code=data.trampoline_cpp_code,
            ifdef=data.ifdef,
            ifndef=data.ifndef,
            release_gil=release_gil,
            template_impls=data.template_impls,
            virtual_xform=data.virtual_xform,
            is_overloaded=overload_tracker,
            _fn=fn,
        )

        # Generate a special lambda wrapper only when needed
        if not data.cpp_code and (has_out_param or fctx.has_buffers):
            self._on_fn_make_lambda(data, fctx)

        return fctx

    def _on_fn_param(
        self,
        p: Parameter,
        p_name: str,
        fn_disable_none: typing.Optional[bool],
        param_override: ParamData,
        param_remap: typing.Dict[str, str],
    ):
        ptype, p_pointer, p_reference, p_const = _count_and_unwrap(p.type)
        fundamental = isinstance(ptype, Type) and _is_fundamental(
            ptype.typename.segments[-1]
        )
        self._add_type_caster(ptype)

        # TODO: get rid of this, use const by default?
        if isinstance(ptype, Type) and ptype.const:
            ptype.const = False
            cpp_type_no_const = ptype.format()
            ptype.const = True
        else:
            cpp_type_no_const = ptype.format()

        cpp_type = cpp_type_no_const

        if p_pointer:
            call_name = p_name
        elif p_reference:
            call_name = f"std::forward<decltype({p_name})>({p_name})"
        else:
            call_name = f"std::move({p_name})"

        # This is different because call_name might get special treatment later
        virtual_call_name = call_name
        cpp_retname = orig_pname = p_name

        # TODO: this is precarious
        # - needs to override some things
        force_out = False
        default = None
        disable_none = fn_disable_none
        if param_override is not _default_param_data:
            force_out = param_override.force_out
            if param_override.name:
                p_name = param_override.name
            if param_override.x_type:
                cpp_type = param_override.x_type
                self._add_user_type_caster(cpp_type)
            if param_override.default:
                default = param_override.default
            if param_override.no_default:
                default = None
            if param_override.disable_none is not None:
                disable_none = param_override.disable_none

        py_pname = p_name
        if iskeyword(py_pname):
            py_pname = f"{py_pname}_"

        if orig_pname != py_pname:
            param_remap[orig_pname] = py_pname

        # Autodetect disable_none if not explicitly specified
        if disable_none is None:
            disable_none = cpp_type.startswith("std::function")

        if disable_none:
            py_arg = f'py::arg("{py_pname}").none(false)'
        else:
            py_arg = f'py::arg("{py_pname}")'

        #
        # Default parameter
        #

        # Do this after cpp_type is resolved but before it gets its const
        if not default and p.default and not param_override.no_default:
            default = tokfmt(p.default.tokens)

        if default:
            default = self._resolve_default(default, cpp_type, ptype)
            if not param_override.disable_type_caster_default_cast:
                default = self._add_default_arg_cast(default, ptype)
            if default:
                py_arg = f"{py_arg} = {default}"

        pcat = ParamCategory.IN

        if force_out or (
            (p_pointer or p_reference == 1) and not p_const and fundamental
        ):
            if p_pointer:
                call_name = f"&{call_name}"
            else:
                call_name = p_name

            pcat = ParamCategory.OUT
        elif isinstance(ptype, Array):
            asz = param_override.array_size
            if not asz:
                asz = _fmt_array_size(ptype)

            if asz is not None:
                cpp_type = f"std::array<{ptype.array_of.format()}, {asz}>"
                call_name = f"{call_name}.data()"
                if not default:
                    default = "{}"
                self._add_user_type_caster("std::array")
            else:
                # it's a vector
                pass
            pcat = ParamCategory.OUT

        if p_const:
            cpp_type = f"const {cpp_type}"

        # TODO: this is weird, why aren't we using .format() here
        x_type_full = cpp_type
        x_type_full += "&" * p_reference
        x_type_full += "*" * p_pointer

        return ParamContext(
            arg_name=p_name,
            cpp_type=cpp_type,
            full_cpp_type=x_type_full,
            py_arg=py_arg,
            default=default,
            # only used by genlambda
            call_name=call_name,
            # only used if virtual, duh
            virtual_call_name=virtual_call_name,
            cpp_retname=cpp_retname,
            category=pcat,
        )

    def _on_fn_make_lambda(self, data: FunctionData, fctx: FunctionContext):
        """
        When we need to transform the C++ function to make it more pythonic, we
        autogenerate a lambda function as part of the wrapper. This is needed:

        * When an 'out' parameter is detected (a pointer receiving a value)
        * When a buffer + size parameter exists (either in or out)
        * When "renaming" a constructor overload to a static method
        """

        # Statements to insert before calling the function
        lambda_pre: typing.List[str] = []

        # If buffer overrides present, apply those to the parameters
        if data.buffers:
            self._apply_buffer_params(data, fctx, lambda_pre)

        in_params: typing.List[ParamContext] = []
        out_params: typing.List[ParamContext] = []
        ret_params: typing.List[typing.Union[_ReturnParamContext, ParamContext]] = []
        tmp_params: typing.List[ParamContext] = []

        #
        # Sort the parameters
        #

        for pctx in fctx.filtered_params:
            if pctx.category == ParamCategory.OUT:
                out_params.append(pctx)
                tmp_params.append(pctx)

            elif pctx.category == ParamCategory.IN:
                in_params.append(pctx)

            elif pctx.category == ParamCategory.TMP:
                tmp_params.append(pctx)

        call_start = ""
        lambda_ret = ""

        # Return values (original return value + any out parameters)
        fn_retval = fctx.cpp_return_type
        if fn_retval and fn_retval != "void":
            call_start = "auto __ret = "
            ret_params = [_ReturnParamContext(cpp_retname="__ret", cpp_type=fn_retval)]
            ret_params.extend(out_params)
        else:
            ret_params.extend(out_params)

        if len(ret_params) == 1 and ret_params[0].cpp_type != "void":
            lambda_ret = f"return {ret_params[0].cpp_retname};"
        elif len(ret_params) > 1:
            t = ",".join([p.cpp_retname for p in ret_params])
            lambda_ret = f"return std::make_tuple({t});"

        # Temporary values to store out parameters in
        if tmp_params:
            for out in reversed(tmp_params):
                odef = out.default
                if not odef:
                    lambda_pre.insert(0, f"{out.cpp_type} {out.arg_name}")
                elif odef.startswith("{"):
                    lambda_pre.insert(0, f"{out.cpp_type} {out.arg_name}{odef}")
                else:
                    lambda_pre.insert(0, f"{out.cpp_type} {out.arg_name} = {odef}")

        pre = f";\n".join(lambda_pre) + ";"

        fctx.genlambda = GeneratedLambda(
            pre=pre,
            call_start=call_start,
            ret=lambda_ret,
            in_params=in_params,
            out_params=out_params,
        )

    def _apply_buffer_params(
        self,
        data: FunctionData,
        fctx: FunctionContext,
        lambda_pre: typing.List[str],
    ):
        """
        Modifies the function parameters for buffer usage
        """

        # buffers: accepts a python object that supports the buffer protocol
        #          as input. If the buffer is an 'out' buffer, then it
        #          will request a writeable buffer. Data is written by the
        #          wrapped function to that buffer directly, and the length
        #          written (if the length is a pointer) will be returned
        buffer_params: typing.Dict[str, BufferData] = {}
        buflen_params: typing.Dict[str, bool] = {}

        for bufinfo in data.buffers:
            if bufinfo.src in buffer_params:
                raise ValueError(
                    f"buffer src({bufinfo.src}) is in multiple buffer specifications"
                )
            # length can be shared
            # elif bufinfo.len in buflen_params:
            #     raise ValueError(
            #         f"buffer len({bufinfo.len}) is in multiple buffer specifications"
            #     )
            buffer_params[bufinfo.src] = bufinfo
            buflen_params[bufinfo.len] = False

        dups = set(buffer_params.keys()).intersection(buflen_params.keys())
        if dups:
            names = "', '".join(dups)
            raise ValueError(f"These params are both buffer src and len: '{names}'")

        for pctx in fctx.all_params:
            p_name = pctx.arg_name
            if p_name in buffer_params:
                bufinfo = buffer_params.pop(p_name)
                bname = f"__{bufinfo.src}"

                pctx.call_name = f"({pctx.cpp_type}*){bname}.ptr"
                pctx.cpp_type = "const py::buffer"
                pctx.full_cpp_type = "const py::buffer&"

                # this doesn't seem to be true for bytearrays, which is silly
                # x_lambda_pre.append(
                #     f'if (PyBuffer_IsContiguous((Py_buffer*){p_name}.ptr(), \'C\') == 0) throw py::value_error("{p_name}: buffer must be contiguous")'
                # )

                # TODO: check for dimensions, strides, other dangerous things

                # bufinfo was validated and converted before it got here
                pctx.category = ParamCategory.IN
                if bufinfo.type is BufferType.IN:
                    lambda_pre += [f"auto {bname} = {p_name}.request(false)"]
                else:
                    lambda_pre += [f"auto {bname} = {p_name}.request(true)"]

                lambda_pre += [f"{bufinfo.len} = {bname}.size * {bname}.itemsize"]

                if bufinfo.minsz:
                    lambda_pre.append(
                        f'if ({bufinfo.len} < {bufinfo.minsz}) throw py::value_error("{p_name}: minimum buffer size is {bufinfo.minsz}")'
                    )

            elif p_name in buflen_params:
                buflen_params[p_name] = True

                # If the length is a pointer, assume that the function will accept
                # an incoming length, and set the outgoing length
                if pctx.full_cpp_type.endswith("*"):
                    pctx.call_name = f"&{p_name}"
                    pctx.category = ParamCategory.OUT
                else:
                    # if it's not a pointer, then the called function
                    # can't communicate through it, so ignore the parameter
                    pctx.call_name = p_name
                    pctx.category = ParamCategory.TMP

        if buffer_params:
            names = "', '".join(buffer_params.keys())
            raise ValueError(f"incorrect buffer param names '{names}'")

        unused_buflen_params = [k for k, v in buflen_params.items() if not v]
        if unused_buflen_params:
            names = "', '".join(unused_buflen_params)
            raise ValueError(f"incorrect buffer length names '{names}'")

    #
    # Utility methods
    #

    def _get_fn_name(self, fn: Function) -> typing.Optional[str]:
        s = fn.name.segments[-1]
        if isinstance(s, NameSpecifier):
            return s.name

        # name is too complicated (can this happen? should be a warning?)
        assert False
        return None

    def _get_module_var(self, data: HasSubpackage) -> str:
        if data.subpackage:
            var = f"pkg_{data.subpackage.replace('.', '_')}"
            self.hctx.subpackages[data.subpackage] = var
            return var

        return "m"

    def _make_py_name(
        self,
        name: str,
        data: HasNameData,
        strip_prefixes: typing.Optional[typing.List[str]] = None,
        is_operator: bool = False,
    ):
        if data.rename:
            return data.rename

        if strip_prefixes is None:
            strip_prefixes = self.user_cfg.strip_prefixes

        if strip_prefixes:
            for pfx in strip_prefixes:
                if name.startswith(pfx):
                    n = name[len(pfx) :]
                    if n.isidentifier():
                        name = n
                        break

        if iskeyword(name):
            return f"{name}_"
        if not name.isidentifier() and not is_operator:
            if not self.report_only:
                raise ValueError(f"name {name!r} is not a valid identifier")

        return name

    def _process_doc(
        self,
        doxygen: typing.Optional[str],
        data: HasDoc,
        append_prefix: str = "",
        param_remap: typing.Dict[str, str] = {},
    ) -> Documentation:
        doc = ""

        if data.doc is not None:
            doc = data.doc
        elif doxygen:
            doc = doxygen
            if param_remap:
                d = sphinxify.Doc.from_comment(doc)
                for param in d.params:
                    new_name = param_remap.get(param.name)
                    if new_name:
                        param.name = new_name
                doc = str(d)
            else:
                doc = sphinxify.process_raw(doc)

        if data.doc_append is not None:
            doc += f"\n{append_prefix}" + data.doc_append.replace(
                "\n", f"\n{append_prefix}"
            )

        return self._quote_doc(doc)

    def _quote_doc(self, doc: typing.Optional[str]) -> Documentation:
        doc_quoted: Documentation = None
        if doc:
            # TODO
            doc = doc.replace("\\", "\\\\").replace('"', '\\"')
            doc_quoted = doc.splitlines(keepends=True)
            doc_quoted = ['"%s"' % (dq.replace("\n", "\\n"),) for dq in doc_quoted]

        return doc_quoted

    def _extract_typealias(
        self,
        in_ta: typing.List[str],
        out_ta: typing.List[str],
        ta_names: typing.Set[str],
    ):
        for typealias in in_ta:
            if typealias.startswith("template"):
                out_ta.append(typealias)
            else:
                teq = typealias.find("=")
                if teq != -1:
                    ta_name = typealias[:teq].strip()
                    out_ta.append(f"using {typealias}")
                else:
                    ta_name = typealias.split("::")[-1]
                    out_ta.append(f"using {ta_name} = {typealias}")
                ta_names.add(ta_name)

    def _resolve_default(
        self,
        name: str,
        cpp_type: str,
        ptype: typing.Union[Array, FunctionType, Type],
    ) -> str:
        if name.isnumeric() or name in ("NULL", "nullptr"):
            pass
        elif name[0] == "{" and name[-1] == "}":
            if isinstance(ptype, Array):
                return name
            return f"{cpp_type}{name}"

        # if there's a parent, look there
        # -> this seems rather expensive for little reward, how often do we need
        #    this? Also, doesn't have any test coverage yet, so let's not do it
        #    for now
        #
        # parent = fn["parent"]
        # if parent:
        #     for prop in parent["properties"]["public"]:
        #         if prop["name"] == name:
        #             name = f"{parent['namespace']}::{parent['name']}::{name}"

        return name

    def _add_default_arg_cast(
        self, name: str, ptype: typing.Union[Array, FunctionType, Type]
    ) -> str:
        # Adds an explicit cast to a default arg for certain types that have
        # a type caster with an explicit default
        ntype: typing.Union[Array, FunctionType, Pointer, Type] = ptype
        while isinstance(ntype, Array):
            ntype = ntype.array_of
        if isinstance(ntype, Type):
            typename = _fmt_nameonly(ntype.typename)
            if typename:
                ccfg = self.casters.get(typename)
                if ccfg and ccfg.get("darg"):
                    found_typename = ccfg["typename"]
                    name = f"({found_typename}){name}"

        return name

    #
    # type caster utilities
    #

    def _add_type_caster(self, t: typing.Union[DecoratedType, FunctionType]):
        # pick apart the type and add each to the list of types
        # - it would be nice if we could just add this to a set
        #   and process it later, but that would probably be just
        #   as much work?
        while True:
            if isinstance(t, Type):
                self._add_type_caster_pqname(t.typename)
                return

            elif isinstance(t, FunctionType):
                self._add_type_caster(t.return_type)
                for p in t.parameters:
                    self._add_type_caster(p.type)
                return

            elif isinstance(t, Pointer):
                t = t.ptr_to
            elif isinstance(t, Reference):
                t = t.ref_to
            elif isinstance(t, MoveReference):
                t = t.moveref_to
            elif isinstance(t, Array):
                t = t.array_of
            else:
                assert False

    def _add_type_caster_pqname(self, typename: PQName):
        parts = []
        for p in typename.segments:
            if not isinstance(p, NameSpecifier):
                return
            parts.append(p.name)
            if p.specialization:
                for a in p.specialization.args:
                    if not isinstance(a.arg, Value):
                        self._add_type_caster(a.arg)

        self.types.add("::".join(parts))

    def _add_user_type_caster(self, typename: str):
        # defer until the end since there's lots of duplication
        self.user_types.add(typename)

    def _process_user_type_casters(self):
        # processes each user type caster and adds it to the processed list
        types = self.types
        for typename in self.user_types:
            tmpl_idx = typename.find("<")
            if tmpl_idx == -1:
                types.add(typename)
            else:
                types.add(typename[:tmpl_idx])
                types.update(
                    _type_caster_seps.split(typename[tmpl_idx:].replace(" ", ""))
                )

    def _set_type_caster_includes(self):
        # process user casters
        self._process_user_type_casters()
        casters = self.casters

        # identify any associated headers
        includes = set()
        for typename in self.types:
            ccfg = casters.get(typename)
            if ccfg:
                includes.add(ccfg["hdr"])

        self.hctx.type_caster_includes = sorted(includes)


def parse_header(
    name: str,
    header_path: pathlib.Path,
    header_root: pathlib.Path,
    gendata: GeneratorData,
    parser_options: ParserOptions,
    casters: typing.Dict[str, typing.Dict[str, typing.Any]],
    report_only: bool,
) -> HeaderContext:
    user_cfg = gendata.data

    # Initialize the header context with user configuration
    hctx = HeaderContext(
        hname=name,
        extra_includes_first=user_cfg.extra_includes_first,
        extra_includes=user_cfg.extra_includes,
        inline_code=user_cfg.inline_code,
        rel_fname=str(header_path.relative_to(header_root)),
    )

    # Parse the header using a custom visitor
    visitor = AutowrapVisitor(hctx, gendata, casters, report_only)
    parser = CxxParser(
        str(header_path), None, visitor, parser_options, encoding=user_cfg.encoding
    )
    parser.parse()

    #
    # Per-header user specified data
    #

    for i, (k, tmpl_data) in enumerate(user_cfg.templates.items()):
        qualname = tmpl_data.qualname
        if "::" not in qualname:
            qualname = f"::{qualname}"
        qualname = qualname.translate(_qualname_trans)

        doc_add = tmpl_data.doc_append
        if doc_add:
            doc_add = f"\n{doc_add}"

        tctx = TemplateInstanceContext(
            scope_var=visitor._get_module_var(tmpl_data),
            var_name=f"tmplCls{i}",
            py_name=k,
            full_cpp_name_identifier=qualname,
            binder_typename=f"bind_{qualname}_{i}",
            params=tmpl_data.params,
            header_name=f"{qualname}.hpp",
            doc_set=visitor._quote_doc(tmpl_data.doc),
            doc_add=visitor._quote_doc(doc_add),
        )
        hctx.template_instances.append(tctx)

        # Ensure that template instances are created in class order if the
        # template class is in this header file
        # - not matching here is not an error
        qualname_match = tmpl_data.qualname.lstrip(":")
        for cctx in hctx.classes:
            if cctx.dep_cpp_name.lstrip(":") == qualname_match:
                assert cctx.template
                tctx.matched = True
                cctx.template.instances.append(tctx)
                break

        for param in tmpl_data.params:
            visitor._add_user_type_caster(param)

    # User typealias additions
    visitor._extract_typealias(user_cfg.typealias, hctx.user_typealias, set())

    # Type caster
    visitor._set_type_caster_includes()

    return hctx



import typing

from .context import FunctionContext
from cxxheaderparser.types import (
    Array,
    DecoratedType,
    FunctionType,
    FundamentalSpecifier,
    Method,
    MoveReference,
    NameSpecifier,
    Pointer,
    Reference,
    Type,
)

# yes, we include some stdint types in here, but that's fine, this
# is just a best effort generator
_builtins = {
    "void": "v",
    "wchar_t": "w",
    "bool": "b",
    "char": "c",
    "signed char": "a",
    "int8_t": "a",
    "unsigned char": "h",
    "uint8_t": "c",
    "short": "s",
    "int16_t": "s",
    "unsigned short": "t",
    "uint16_t": "t",
    "int": "i",
    "int32_t": "i",
    "unsigned int": "j",
    "uint32_t": "j",
    "long": "l",
    "unsigned long": "m",
    "long long": "x",
    "__int64": "x",
    "int64_t": "x",
    "unsigned long long": "y",
    "unsigned __int64": "y",
    "uint64_t": "y",
    "__int128": "n",
    "unsigned __int128": "o",
    "float": "f",
    "double": "d",
    "long double": "e",
    "__float128": "g",
    "char32_t": "Di",
    "char16_t": "Ds",
    "auto": "Da",
}

_type_bad_chars = ":<>=()&,"
_type_trans = str.maketrans(_type_bad_chars, "_" * len(_type_bad_chars))


def _encode_type(dt: DecoratedType, names: typing.List[str]) -> str:
    # Decode the type
    ptrs = 0
    refs = 0
    const = False
    volatile = False

    t = dt
    while True:
        if isinstance(t, Type):
            const = const or t.const
            volatile = volatile or t.volatile
            break
        elif isinstance(t, Pointer):
            ptrs += 1
            const = const or t.const
            volatile = volatile or t.volatile
            t = t.ptr_to
        elif isinstance(t, Reference):
            refs += 1
            t = t.ref_to
        elif isinstance(t, MoveReference):
            refs += 2
            t = t.moveref_to
        else:
            break

    # prefix with cv-qualifiers, refs, pointers
    if const:
        names.append("K")
    if volatile:
        names.append("V")

    if isinstance(t, Array):
        assert False  # TODO, convert array size?
        names.append("A" * t.size)

    if refs == 1:
        names.append("R")
    elif refs == 2:
        names.append("O")

    if ptrs:
        names.append("P" * ptrs)

    if isinstance(t, FunctionType):
        # encode like a function but include the return type
        names.append("F")
        _encode_type(t.return_type, names)
        params = t.parameters
        if not params:
            names.append("_v")
        else:
            for p in params:
                names.append("_")
                _encode_type(p.type, names)
        if t.vararg:
            names.append("_z")
    else:
        typename = t.typename.segments[-1]
        if isinstance(typename, (FundamentalSpecifier, NameSpecifier)):
            typ = _builtins.get(typename.name)
            if not typ:
                # .. good enough, there are cases where this would fail, but
                # hopefully that doesn't happen?
                typ = f"T{typename.name.translate(_type_trans)}"
        else:
            typ = "T__"

        names.append(typ)


def trampoline_signature(fctx: FunctionContext) -> str:
    """
    In our trampoline functions, each function can be disabled by defining
    a macro corresponding to the function type. This helper function
    generates a mangled name based on the function name and types.

    This is roughly based on the IA64 name mangling scheme that GCC uses,
    but omits template stuff.

    function name
    cv qualifiers
    parameter types
    """

    # fast path in case it was computed previously
    if fctx._trampoline_signature:
        return fctx._trampoline_signature

    # TODO: operator overloads
    names = []

    fn = fctx._fn
    if isinstance(fn, Method):
        if fn.const:
            names.append("K")
        refqual = fn.ref_qualifier
        if refqual:
            if refqual == "&":
                names.append("R")
            if refqual == "&&":
                names.append("O")

    names.append(fctx.cpp_name)

    params = fn.parameters
    if not params:
        names.append("_v")
    else:
        for p in params:
            names.append("_")
            _encode_type(p.type, names)

    if fn.vararg:
        names.append("_z")

    fctx._trampoline_signature = "".join(names)
    return fctx._trampoline_signature



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import pkg_resources

os.environ["GENERATING_DOCUMENTATION"] = "True"

__version__ = pkg_resources.get_distribution("robotpy-build").version


# -- RTD configuration ------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# -- Project information -----------------------------------------------------

project = "robotpy-build"
copyright = "2020, RobotPy Development Team"
author = "RobotPy Development Team"

# The full version, including alpha/beta/rc tags
version = ".".join(__version__.split(".")[:3])
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
]

# For sphinx-autodoc-typehints
always_document_param_types = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "pybind11": (
        "https://pybind11.readthedocs.io/en/latest/",
        None,
    )
}


# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



import json
import os
from os.path import join
import pprint
import typing

from .context import HeaderContext
from .render_wrapped import render_wrapped_cpp
from .render_cls_rpy_include import render_cls_rpy_include_hpp
from .render_tmpl_inst import render_template_inst_cpp, render_template_inst_hpp

_emit_j2_debug = os.getenv("RPYBUILD_J2_DEBUG") == "1"


class WrapperWriter:

    def write_files(
        self,
        hctx: HeaderContext,
        name: str,
        cxx_gen_dir: str,
        hppoutdir: str,
        classdeps_json_fname: str,
    ) -> typing.List[str]:
        """Generates all files needed for a single processed header"""

        generated_sources: typing.List[str] = []

        # Jinja requires input as a dictionary
        data = hctx.__dict__

        if _emit_j2_debug:
            with open(join(cxx_gen_dir, f"{name}.txt"), "w") as fp:
                fp.write(pprint.pformat(hctx))

        # Write the cpp file first
        fname = join(cxx_gen_dir, f"{name}.cpp")
        generated_sources.append(fname)
        with open(fname, "w", encoding="utf-8") as fp:
            fp.write(render_wrapped_cpp(hctx))

        # Then the json
        with open(classdeps_json_fname, "w", encoding="utf-8") as fp:
            json.dump(hctx.class_hierarchy, fp)

        # Generate an rpy-include file for each class that has either a trampoline
        # or a template class
        for cls in hctx.classes:
            if not cls.template and not cls.trampoline:
                continue

            fname = join(
                hppoutdir, f"{cls.namespace.replace(':', '_')}__{cls.cpp_name}.hpp"
            )
            with open(fname, "w", encoding="utf-8") as fp:
                fp.write(render_cls_rpy_include_hpp(hctx, cls))

        # Each class template is instantiated in a separate cpp file to lessen
        # compiler memory requirements when compiling obnoxious templates
        if hctx.template_instances:
            # Single header output that holds all the struct outlines
            fname = join(cxx_gen_dir, f"{name}_tmpl.hpp")
            with open(fname, "w", encoding="utf-8") as fp:
                fp.write(render_template_inst_hpp(hctx))

            # Each cpp file has a single class template instance
            for i, tmpl_data in enumerate(hctx.template_instances):
                data["tmpl_data"] = tmpl_data
                fname = join(hppoutdir, f"{name}_tmpl{i+1}.cpp")
                generated_sources.append(fname)
                with open(fname, "w", encoding="utf-8") as fp:
                    fp.write(render_template_inst_cpp(hctx, tmpl_data))

        return generated_sources



from rpytest import ft


def test_overloaded():
    class PyOB(ft.OB):
        def init(self, init):
            return f"PyOB => {init.doInit()}"

    class PyOC(ft.OC):
        def init(self, init):
            return f"PyOC => {init.doInit()}"

    class PyOG(ft.OG):
        def init(self, init):
            return f"PyOG => {init.doInit()}"

    obinit = ft.OBInitializer()
    ocinit = ft.OCInitializer()

    ob = PyOB()
    assert ob.init(obinit) == "PyOB => OBInitializer"
    assert ft.OBinitOB(ob, obinit) == "PyOB => OBInitializer"
    assert ft.OBinitOB(ob, ocinit) == "PyOB => OCInitializer"

    oc = PyOC()
    assert oc.init(obinit) == "PyOC => OBInitializer"
    assert oc.init(ocinit) == "PyOC => OCInitializer"
    assert ft.OCinitOB(oc, obinit) == "PyOC => OBInitializer"
    assert ft.OCinitOC(oc, obinit) == "PyOC => OBInitializer"
    assert ft.OCinitOC(oc, ocinit) == "PyOC => OCInitializer"

    og = PyOG()
    assert og.init(obinit) == "PyOG => OBInitializer"
    assert og.init(ocinit) == "PyOG => OCInitializer"
    assert ft.OCinitOB(og, obinit) == "PyOG => OBInitializer"
    assert ft.OCinitOC(og, obinit) == "PyOG => OBInitializer"
    assert ft.OCinitOC(og, ocinit) == "PyOG => OCInitializer"
    # assert ft.OGinitOB(og, obinit) == "PyOG => OBInitializer"
    assert ft.OGinitOC(og, ocinit) == "PyOG => OCInitializer"

    og = ft.OG()
    # assert og.init(obinit) == "" # scope resolution would fail in C++, won't work here either
    assert og.init(ocinit) == "OG::Init(OCInitializer &init) => OCInitializer"



#
# These dataclasses hold data to be rendered
#
# To make the rendering functions easier to understand, where possible we try
# to do logic in the code that produces this data instead of in the rendering
# functions.
#
# We also prefer to copy over data from the autowrap YAML file instead of using
# those data structures directly. While there's some slight overhead added,
# this should help to keep the logic outside of the templates.
#

from dataclasses import dataclass, field
import enum
import typing

from cxxheaderparser.types import Function, PQName

from ..config.autowrap_yml import ReturnValuePolicy


Documentation = typing.Optional[typing.List[str]]


class OverloadTracker:
    """Evaluates to true if overloaded"""

    def __init__(self) -> None:
        self.overloads = 0

    def add_overload(self):
        # This only needs to count new overloads, because if there are
        # multiple instances of a function it *has* to be different or
        # the compiler would complain
        self.overloads += 1

    def __bool__(self) -> bool:
        return self.overloads > 1


@dataclass
class EnumeratorContext:
    """Render data for each enumerator"""

    #: C++ name, including namespace/classname
    full_cpp_name: str

    #: Name in python
    py_name: str

    #: Documentation
    doc: Documentation


@dataclass
class EnumContext:
    """Render data for enum"""

    #: Name of parent variable in initializer
    scope_var: str

    #: Name of variable in initializer
    var_name: str

    #: C++ name without namespace
    cpp_name: str

    #: C++ name, including namespace/classname
    full_cpp_name: str

    #: Python name
    py_name: str

    #: Enum values
    values: typing.List[EnumeratorContext]

    #: Documentation
    doc: Documentation

    #
    # Copied from user's EnumData
    #

    arithmetic: bool
    inline_code: typing.Optional[str]


class ParamCategory(enum.Enum):
    IGNORE = 0
    OUT = 1
    IN = 2
    TMP = 3


@dataclass
class ParamContext:
    """Render data for each parameter"""

    # Original argument name
    arg_name: str

    # name of type with const but no *, &
    # .. why does this have const
    cpp_type: str

    # contains 'const', &, etc
    full_cpp_type: str

    #: py::arg() for pybind11, includes default value
    py_arg: str

    #: passed to lambda
    default: typing.Optional[str]

    #: Name to pass to function when calling the original
    #: .. only used by lambda
    call_name: str

    #: Name to pass to function when generating a virtual function
    virtual_call_name: str

    #: Not used in renderer, but is used to determine variable
    #: name when used as an out parameter
    cpp_retname: str

    #: Not used in renderer, used for filtering
    category: ParamCategory

    # type + name, rarely used
    @property
    def decl(self) -> str:
        return f"{self.full_cpp_type} {self.arg_name}"

    # only used for operator generation, rarely used
    @property
    def cpp_type_no_const(self) -> str:
        ct = self.cpp_type
        if ct.startswith("const "):
            return ct[6:]
        return ct


@dataclass
class GeneratedLambda:
    """
    Data for generating a lambda to change the behavior of the function
    """

    pre: str
    call_start: str
    ret: str

    #: input parameters
    in_params: typing.List[ParamContext]
    #: output parameters
    out_params: typing.List[ParamContext]


@dataclass
class FunctionContext:
    """Render data for a C++ function or method"""

    #: C++ name of function
    cpp_name: str

    #: Documentation
    doc: Documentation

    #: parent variable to attach to
    scope_var: str

    #: Name in python
    py_name: str

    #: Return value as fully qualified C++ type with const/*
    cpp_return_type: typing.Optional[str]

    #: All original C++ parameters for the function
    #: -> used by trampoline signature, and trampolines
    all_params: typing.List[ParamContext]

    #: every parameter except ignored
    filtered_params: typing.List[ParamContext]

    #: Has vararg parameters
    #: -> used by trampoline signature
    # vararg: bool

    #
    # Mixed
    #

    has_buffers: bool

    keepalives: typing.List[typing.Tuple[int, int]]

    return_value_policy: str

    #
    # User settings from autowrap_yml.FunctionData
    #

    #: If True, don't wrap this, but provide a pure virtual implementation
    ignore_pure: bool

    #: If True, don't expose this function to python
    ignore_py: bool

    #: Use this code instead of the generated code
    cpp_code: typing.Optional[str]
    #: Use this code instead of the generated code in a trampoline
    trampoline_cpp_code: typing.Optional[str]

    #: Generate this in an `#ifdef`
    ifdef: typing.Optional[str]
    #: Generate this in an `#ifndef`
    ifndef: typing.Optional[str]

    release_gil: bool

    # List of template instantiations
    template_impls: typing.Optional[typing.List[typing.List[str]]]

    virtual_xform: typing.Optional[str]

    # OverloadTracker evaluates to True if there are overloads
    is_overloaded: OverloadTracker

    # Used to compute the trampoline signature
    _fn: Function

    #
    # Cached/conditionally set properties
    #

    genlambda: typing.Optional[GeneratedLambda] = None

    #: Is this a constructor?
    is_constructor: bool = False

    #: & or && qualifiers for function
    ref_qualifiers: str = ""

    #: Marked const
    const: bool = False

    is_pure_virtual: bool = False

    #: If there is a namespace associated with this function, this is it,
    #: and this ends with ::
    namespace: str = ""

    #: The operator for this method
    #: - if set, cpp_type will be filled out by the parser
    operator: typing.Optional[str] = None

    #: True if this is a static method
    is_static_method: bool = False

    # Only compute the trampoline signature once, used as cache by
    # trampoline_signature function
    _trampoline_signature: typing.Optional[str] = None


@dataclass
class PropContext:
    """
    Render data for each class property
    """

    py_name: str
    cpp_name: str
    cpp_type: str
    readonly: bool
    doc: Documentation

    array_size: typing.Optional[int]
    array: bool  # cannot sensibly autowrap an array of incomplete size
    reference: bool
    static: bool
    bitfield: bool


@dataclass
class BaseClassData:
    """
    Render data for each base that a class inherits
    """

    #: Just the class name
    cls_name: str

    #: This ends with ::
    namespace_: str

    #: C++ name, including all known components
    full_cpp_name: str  # was x_qualname

    full_cpp_name_w_templates: str  # was x_class

    #: Translated C++ name suitable for use as an identifier. :<>= are
    #: turned into underscores.
    full_cpp_name_identifier: str  # was x_qualname_

    #: C++ name + components, no template parameters
    dep_cpp_name: str

    #: comma separated list of template parameters for this base, or empty string
    template_params: str


@dataclass
class TrampolineData:
    """
    Trampolines are classes that have the original class as a base class,
    which allows us to add additional features from python:

    * Protected method/variable access
    * Implement virtual functions from python
    """

    full_cpp_name: str
    var: str
    inline_code: typing.Optional[str]

    #: trampoline base class template types
    tmpl_args: str

    #: trampoline base class template names
    tmpl_params: str

    #
    # Lists used by trampolines so that the logic in the template is less
    # complicated
    #

    # pub + protected + final (ignore doesn't matter)
    # private + final or override (ignore doesn't matter)
    # -> to delete, only needs signature
    methods_to_disable: typing.List[FunctionContext]

    # public + protected + private
    # - if not ignore and virtual and not final and not buffers)
    virtual_methods: typing.List[FunctionContext]

    # protected + not ignore + constructor
    protected_constructors: typing.List[FunctionContext]

    # protected + not (ignore or virtual or constructor)
    non_virtual_protected_methods: typing.List[FunctionContext]


@dataclass
class ClassTemplateData:
    #: N, ..
    argument_list: str
    #: <typename N, .. >
    parameter_list: str

    #: the specified C++ code is inserted into the template definition
    inline_code: str

    #: Instances of this class
    instances: typing.List["TemplateInstanceContext"] = field(default_factory=list)


@dataclass
class ClassContext:
    """
    Render data for each class encountered in a header
    """

    parent: typing.Optional["ClassContext"]

    #: Namespace that this class lives in
    namespace: str

    #: C++ name (only the class)
    cpp_name: str

    #: C++ name, including namespace/classname: was: x_qualname
    full_cpp_name: str

    #: Translated C++ name suitable for use as an identifier. :<>= are
    #: turned into underscores. was: x_qualname_
    full_cpp_name_identifier: str

    #: C++ name + components, no template parameters
    dep_cpp_name: str

    #: Python name
    py_name: str

    #: Name of parent variable in initializer
    scope_var: str

    #: Name of variable in initializer. was: x_varname
    var_name: str

    #: If the object shouldn't be deleted by pybind11, use this. Disables
    #: implicit constructors.
    nodelete: bool

    #: class is final
    final: bool

    #: Documentation
    doc: Documentation

    bases: typing.List[BaseClassData]

    template: typing.Optional[ClassTemplateData]

    #
    # User specified settings copied from ClassData
    #

    #: Extra 'using' directives to insert into the trampoline and the
    #: wrapping scope
    user_typealias: typing.List[str]

    #: Extra constexpr to insert into the trampoline and wrapping scopes
    #: (name, value)
    constants: typing.List[typing.Tuple[str, str]]

    #: Extra code to insert into the class scope
    inline_code: str

    #: User specified settings
    force_multiple_inheritance: bool

    #
    # Everything else
    #

    # Not used in rendering functions but used elsewhere
    has_constructor: bool = False
    is_polymorphic: bool = False

    #: was x_has_trampoline
    trampoline: typing.Optional[TrampolineData] = None

    #
    # Properties (member variables)
    #

    public_properties: typing.List[PropContext] = field(default_factory=list)
    protected_properties: typing.List[PropContext] = field(default_factory=list)

    #
    # Methods: the idea here is have a bunch of descriptive lists here so that
    # the j2 templates don't need logic to emit each method
    #

    #
    # Method lists for wrapping
    #

    add_default_constructor: bool = False

    # public + not (ignore_pure + ignore_py)
    wrapped_public_methods: typing.List[FunctionContext] = field(default_factory=list)

    # only if trampoline:
    # - protected + not (ignore_pure + ignore_py)
    wrapped_protected_methods: typing.List[FunctionContext] = field(
        default_factory=list
    )

    #: Public enums + unnamed enums
    enums: typing.List[EnumContext] = field(default_factory=list)
    unnamed_enums: typing.List[EnumContext] = field(default_factory=list)

    #: Extra autodetected 'using' directives
    auto_typealias: typing.List[str] = field(default_factory=list)

    #: vcheck are various static asserts that check things about the
    #: inline functions
    vcheck_fns: typing.List[FunctionContext] = field(default_factory=list)

    child_classes: typing.List["ClassContext"] = field(default_factory=list)


@dataclass
class TemplateInstanceContext:
    #: Name of parent variable in initializer
    scope_var: str

    #: Name of variable in initializer
    var_name: str

    py_name: str

    full_cpp_name_identifier: str
    binder_typename: str

    params: typing.List[str]

    header_name: str

    doc_set: Documentation
    doc_add: Documentation

    #: If true, instantiated in class order
    matched: bool = False


@dataclass
class HeaderContext:
    """
    Data derived from parsing a single header
    """

    # Name in toml
    hname: str

    extra_includes_first: typing.List[str]
    extra_includes: typing.List[str]
    inline_code: typing.Optional[str]

    #: Path to the parsed header relative to some root
    rel_fname: str

    #: True if <pybind11/operators.h> is needed
    need_operators_h: bool = False

    using_declarations: typing.List[PQName] = field(default_factory=list)

    # TODO: anon enums?
    enums: typing.List[EnumContext] = field(default_factory=list)

    # classes that are not contained in other classes
    classes: typing.List[ClassContext] = field(default_factory=list)

    # same as classes, but only those that have trampolines
    classes_with_trampolines: typing.List[ClassContext] = field(default_factory=list)

    functions: typing.List[FunctionContext] = field(default_factory=list)

    # trampolines

    # template_classes
    template_instances: typing.List[TemplateInstanceContext] = field(
        default_factory=list
    )

    type_caster_includes: typing.List[str] = field(default_factory=list)
    user_typealias: typing.List[str] = field(default_factory=list)

    using_ns: typing.List[str] = field(default_factory=list)

    # All namespaces that occur in the file
    namespaces: typing.List[str] = field(default_factory=list)

    subpackages: typing.Dict[str, str] = field(default_factory=dict)

    # key: class name, value: list of classes this class depends on
    class_hierarchy: typing.Dict[str, typing.List[str]] = field(default_factory=dict)

    has_vcheck: bool = False



#!/usr/bin/env python3

import http.server
import os
from os.path import abspath, dirname, join
import pathlib
import shutil
import sys
import subprocess
import threading
import tempfile
import zipfile


def create_artifact_path(path, group_id, artifact_id, version, classifier):
    components = group_id.split(".") + [artifact_id, version]
    path = join(path, *components)
    os.makedirs(path, exist_ok=True)
    fname = f"{artifact_id}-{version}-{classifier}.zip"
    return join(path, fname)


def http_server():
    httpd = http.server.HTTPServer(
        ("127.0.0.1", 0), http.server.SimpleHTTPRequestHandler
    )

    t = threading.Thread(target=httpd.serve_forever, daemon=True)
    t.start()

    return httpd, httpd.socket.getsockname()[1]


if __name__ == "__main__":
    root = abspath(dirname(__file__))
    os.chdir(root)

    # delete build/cache directory
    shutil.rmtree(join(root, "build", "cache"), ignore_errors=True)

    # create tempdir with maven directory structure for pkg
    with tempfile.TemporaryDirectory() as d:
        # create headers and sources zip files
        hname = create_artifact_path(d, "fake.dl", "dl", "1.2.3", "headers")
        with zipfile.ZipFile(hname, "w") as z:
            z.write(join(root, "dl", "downloaded.h"), "downloaded.h")

        sname = create_artifact_path(d, "fake.dl", "dl", "1.2.3", "sources")
        with zipfile.ZipFile(sname, "w") as z:
            z.write(join(root, "dl", "downloaded.cpp"), "downloaded.cpp")

        # http.server prior to 3.7 could only serve the current directory
        os.chdir(d)

        # start http server on random port
        httpd, port = http_server()

        with open(join(root, "pyproject.toml.tmpl")) as rfp:
            content = rfp.read().replace("RANDOM_PORT", str(port))
            with open(join(root, "pyproject.toml"), "w") as wfp:
                wfp.write(content)

        cwd = None

        if len(sys.argv) == 2 and sys.argv[1] == "wheel":
            cmd_args = [sys.executable, "-m", "build", "--wheel", "--no-isolation"]
            cwd = root
        elif len(sys.argv) == 2 and sys.argv[1] == "develop":
            cmd_args = [sys.executable, "setup.py", "develop", "-N"]
            cwd = root
        else:
            # run pip install
            cmd_args = [
                sys.executable,
                "-m",
                "pip",
                "-v",
                "--disable-pip-version-check",
                "install",
                "--no-build-isolation",
            ]

            if len(sys.argv) == 2 and sys.argv[1] == "-e":
                cmd_args.append("-e")

            cmd_args.append(root)

        env = os.environ.copy()
        env["SETUPTOOLS_SCM_PRETEND_VERSION"] = "0.0.1"

        subprocess.check_call(cmd_args, cwd=cwd, env=env)

        # Windows fails if you try to delete the directory you're currently in
        os.chdir(root)



# Used by robotpybuild entry point

from os.path import abspath, join, dirname
from typing import Any, Dict, List, Optional

_root = abspath(dirname(__file__))


def get_include_dirs() -> Optional[List[str]]:
    return [join(_root, "pybind11", "include"), join(_root, "include")]


def get_library_dirs() -> Optional[List[str]]:
    pass


def get_type_casters_cfg(casters: Dict[str, Dict[str, Any]]) -> None:
    casters.update(
        {
            # STL support
            "std::vector": {"hdr": "pybind11/stl.h"},
            "std::deque": {"hdr": "pybind11/stl.h"},
            "std::list": {"hdr": "pybind11/stl.h"},
            "std::array": {"hdr": "pybind11/stl.h"},
            "std::valarray": {"hdr": "pybind11/stl.h"},
            "std::set": {"hdr": "pybind11/stl.h"},
            "std::map": {"hdr": "pybind11/stl.h"},
            "std::unordered_map": {"hdr": "pybind11/stl.h"},
            "std::optional": {"hdr": "pybind11/stl.h"},
            "std::nullopt_t": {"hdr": "pybind11/stl.h"},
            "std::variant": {"hdr": "pybind11/stl.h"},
            "std::function": {"hdr": "pybind11/functional.h"},
            "std::complex": {"hdr": "pybind11/complex.h"},
            "std::chrono::duration": {"hdr": "pybind11/chrono.h"},
            "std::chrono::time_point": {"hdr": "pybind11/chrono.h"},
            # Eigen support (requires numpy)
            "Eigen::Block": {"hdr": "pybind11/eigen.h"},
            "Eigen::DiagonalMatrix": {"hdr": "pybind11/eigen.h"},
            "Eigen::MatrixBase": {"hdr": "pybind11/eigen.h"},
            "Eigen::Matrix": {"hdr": "pybind11/eigen.h"},
            "Eigen::Matrix2d": {"hdr": "pybind11/eigen.h"},
            "Eigen::Matrix3d": {"hdr": "pybind11/eigen.h"},
            "Eigen::MatrixXcd": {"hdr": "pybind11/eigen.h"},
            "Eigen::MatrixXd": {"hdr": "pybind11/eigen.h"},
            "Eigen::MatrixXdR": {"hdr": "pybind11/eigen.h"},
            "Eigen::MatrixXi": {"hdr": "pybind11/eigen.h"},
            "Eigen::MatrixXf": {"hdr": "pybind11/eigen.h"},
            "Eigen::Ref": {"hdr": "pybind11/eigen.h"},
            "Eigen::Matrix4d": {"hdr": "pybind11/eigen.h"},
            "Eigen::RowVectorXf": {"hdr": "pybind11/eigen.h"},
            "Eigen::SparseMatrix": {"hdr": "pybind11/eigen.h"},
            "Eigen::SparseView": {"hdr": "pybind11/eigen.h"},
            "Eigen::Vector": {"hdr": "pybind11/eigen.h"},
            "Eigen::Vector2d": {"hdr": "pybind11/eigen.h"},
            "Eigen::Vector3d": {"hdr": "pybind11/eigen.h"},
            "Eigen::VectorXf": {"hdr": "pybind11/eigen.h"},
            "Eigen::VectorXcf": {"hdr": "pybind11/eigen.h"},
        }
    )


def get_type_casters(casters: Dict[str, str]) -> None:
    t = {}
    get_type_casters_cfg(t)
    for k, v in t.items():
        if "hdr" in v:
            casters[k] = v["hdr"]



import gc
import sys

from rpytest import ft


def test_ft_autokeepalive():
    # robotpy-build should keepalive the reference without being told in the YML

    p = ft.PatientRef()

    n = ft.Nurse(p)
    assert p is n.m_p

    del p
    gc.collect()

    assert not n.patientDead()



from os.path import abspath
from setuptools.command.editable_wheel import editable_wheel


class EditableWheel(editable_wheel):
    def run(self):
        # you aren't supposed to do this, but... they broke my workflow
        self.distribution.rpybuild_develop_path = abspath(self.project_dir)
        editable_wheel.run(self)



import os
from pydantic import BaseModel

# Needed because pydantic gets in the way of generating good docs
_generating_documentation = bool(os.environ.get("GENERATING_DOCUMENTATION"))
if _generating_documentation:
    BaseModel = object


class Model(BaseModel):
    class Config:
        extra = "forbid"



from .buffer import RenderBuffer
from .context import HeaderContext, TemplateInstanceContext

from .render_cls_prologue import render_class_prologue


def render_template_inst_cpp(
    hctx: HeaderContext, tmpl_data: TemplateInstanceContext
) -> str:
    r = RenderBuffer()
    render_class_prologue(r, hctx)

    tmpl_params = ", ".join(tmpl_data.params)

    r.write_trim(
        f"""
        #include <rpygen/{ tmpl_data.header_name }>
        #include "{ hctx.hname }_tmpl.hpp"

        namespace rpygen {{
        
        using BindType = rpygen::bind_{ tmpl_data.full_cpp_name_identifier }<{tmpl_params}>;
        static std::unique_ptr<BindType> inst;

        { tmpl_data.binder_typename }::{ tmpl_data.binder_typename }(py::module &m, const char * clsName)
        {{
          inst = std::make_unique<BindType>(m, clsName);
        }}

        void { tmpl_data.binder_typename }::finish(const char *set_doc, const char *add_doc)
        {{
          inst->finish(set_doc, add_doc);
          inst.reset();
        }}

        }}; // namespace rpygen
    """
    )
    r.writeln()
    return r.getvalue()


def render_template_inst_hpp(hctx: HeaderContext) -> str:
    r = RenderBuffer()
    r.write_trim(
        f"""
        // This file is autogenerated. DO NOT EDIT
        #pragma once
        #include <robotpy_build.h>

        namespace rpygen {{         
    """
    )

    for tmpl_data in hctx.template_instances:
        r.writeln()
        r.write_trim(
            f"""
            struct {tmpl_data.binder_typename} {{
              {tmpl_data.binder_typename}(py::module &m, const char * clsName);
              void finish(const char *set_doc, const char *add_doc);
            }};
        """
        )

    r.writeln("\n}; // namespace rpygen")
    return r.getvalue()



from rpytest import ft
import pytest


def test_enums():
    # normal enums convert to integers
    assert ft.GEnum.GE1 == 1
    assert ft.NSGEnum.NSGE2 == 2
    assert ft.EnumContainer.InnerEnum.IE1 == 1
    assert ft.NSEnumContainer.InnerEnum.NSIE1 == 1
    assert ft.NSEnumContainer2.InnerEnumContainer.MoreInnerEnum.NSIEMIE1 == 1

    # Unnamed enums are hoisted as integers to their scope
    # - not supported yet for globals
    # assert ft._rpytest_ft.UGEX == 7
    # assert ft._rpytest_ft.NSUGEX == 5
    assert ft.EnumContainer.UEX == 4

    # enum class are specific types
    assert "GCE1" in ft.GCEnum.GCE1.__members__
    assert "NSGE2" in ft.NSGCEnum.NSGE2.__members__
    assert "IEC1" in ft.EnumContainer.InnerCEnum.__members__
    assert "NSIEC1" in ft.NSEnumContainer.InnerCEnum.__members__
    assert (
        "NSIEMIEC1" in ft.NSEnumContainer2.InnerEnumContainer.MoreInnerCEnum.__members__
    )


def test_enum_strip_prefix():
    # 1 isn't a valid identifier, so it's not stripped
    assert ft.StripPrefixEnum.STRIP_1 == 1
    # STRIP_B is converted to B
    assert ft.StripPrefixEnum.B == 2


def test_enum_math():
    assert ft.GEnumMath.MGE1 & 1 == 1
    assert ft.EnumContainer2.InnerMathEnum.IME1 & 1 == 1

    # Normal enums don't support math
    with pytest.raises(TypeError):
        ft.EnumContainer.InnerEnum.IE1 & 1



import os
from os.path import join
import posixpath
import shutil
from typing import Any, Dict, List, Optional

from .download import download_and_extract_zip
from .config.pyproject_toml import Download, StaticLibConfig


class StaticLib:
    # implements pkgcfg

    def __init__(self, name: str, cfg: StaticLibConfig, setup):
        self.package_name = name
        self.name = name
        self.cfg = cfg
        self.static_lib = True
        self.libinit_import = None
        self.pypi_package = None
        # TODO
        self.depends = []

        self.platform = setup.platform

        self.root: Optional[os.PathLike] = None
        self.incdir: Optional[str] = None
        self.libdir: Optional[str] = None

    def set_root(self, root: os.PathLike) -> None:
        self.root = root
        self.libdir = join(self.root, self.name, "lib")
        self.incdir = join(self.root, self.name, "include")

    def get_include_dirs(self) -> Optional[List[str]]:
        if self.incdir is None:
            return

        includes = [self.incdir]
        if self.cfg.download:
            for dl in self.cfg.download:
                if dl.extra_includes:
                    includes += [join(self.incdir, inc) for inc in dl.extra_includes]
        return includes

    def get_library_dirs(self) -> Optional[List[str]]:
        if self.libdir:
            return [self.libdir]

    def get_library_dirs_rel(self) -> Optional[List[str]]:
        pass

    def get_library_names(self) -> Optional[List[str]]:
        # don't do this except on Windows
        if self.platform.os != "windows":
            return

        return self._get_libnames(useext=False)

    def get_library_full_names(self) -> Optional[List[str]]:
        pass

    def get_extra_objects(self) -> Optional[List[str]]:
        if self.platform.os == "windows":
            return

        if self.libdir:
            return [join(self.libdir, lib) for lib in self._get_libnames()]

    def get_type_casters_cfg(self, casters: Dict[str, Dict[str, Any]]) -> None:
        pass

    def _get_dl_libnames(self, dl: Download, useext=True):
        ext = ""
        if useext:
            ext = self.platform.staticext
        return [f"{self.platform.libprefix}{lib}{ext}" for lib in dl.libs]

    def _get_libnames(self, useext=True):
        libs = []
        for dl in self.cfg.download:
            if dl.libs:
                libs += self._get_dl_libnames(dl, useext)
        return libs

    def on_build_dl(self, cache: str, libdir: str):
        self.set_root(libdir)

        shutil.rmtree(self.libdir, ignore_errors=True)
        shutil.rmtree(self.incdir, ignore_errors=True)

        os.makedirs(self.libdir)

        for dl in self.cfg.download:
            if dl.sources is not None:
                raise ValueError(f"{dl.url}: cannot specify sources in static lib")

            if dl.libs is None:
                if dl.incdir is None:
                    raise ValueError(f"{dl.url}: must specify libs in static lib")
                to = {}
            else:
                to = {
                    posixpath.join(dl.libdir, libname): join(self.libdir, libname)
                    for libname in self._get_dl_libnames(dl)
                }

            if dl.incdir is not None:
                to[dl.incdir] = self.incdir

            if dl.dlopenlibs is not None:
                raise ValueError(f"{dl.url}: cannot specify dlopenlibs in static lib")

            download_and_extract_zip(dl.url, to, cache)



from rpytest import ft


def test_rename_fn():
    assert not hasattr(ft._rpytest_ft, "fnOriginal")
    assert ft.fnRenamed() == 0x1
    assert ft.fnRenamedParam(y=4) == 4


def test_rename_cls():
    assert not hasattr(ft._rpytest_ft, "OriginalClass")

    c = ft.RenamedClass()

    assert not hasattr(c, "fnOriginal")
    assert c.fnRenamed() == 0x2
    assert c.fnRenamedParam(y=7) == 7

    assert not hasattr(c, "originalProp")
    assert c.renamedProp == 8
    assert c.getProp() == 8

    c.setProp(9)
    assert c.renamedProp == 9
    assert c.getProp() == 9
    c.renamedProp = 10
    assert c.getProp() == 10

    assert not hasattr(c, "ClassOriginalEnum")
    assert not hasattr(c.ClassRenamedEnum, "Param1")
    assert c.ClassRenamedEnum.P1 == 1

    n = ft.RenamedClass.new(1)
    assert isinstance(n, ft.RenamedClass)
    assert n.renamedProp == 1


def test_rename_enums():
    assert not hasattr(ft._rpytest_ft, "OriginalEnum")

    assert not hasattr(ft.RenamedEnum, "Original1")
    assert ft.RenamedEnum.Renamed1 == 1



from distutils.core import Command
from typing import List
import os.path

from ..wrapper import Wrapper


class BuildGen(Command):
    command_name = "build_gen"
    description = "Generates source files"
    user_options = [
        ("build-base=", "b", "base directory for build library"),
        ("build-temp=", "t", "temporary build directory"),
        ("cxx-gen-dir=", "b", "Directory to write generated C++ files"),
    ]
    wrappers: List[Wrapper] = []

    def initialize_options(self):
        self.build_base = None
        self.build_temp = None
        self.cxx_gen_dir = None

    def finalize_options(self):
        self.set_undefined_options(
            "build", ("build_base", "build_base"), ("build_temp", "build_temp")
        )
        if self.cxx_gen_dir is None:
            self.cxx_gen_dir = os.path.join(self.build_temp, "gensrc")

    def run(self):
        # files need to be downloaded before building can occur
        self.run_command("build_dl")

        for wrapper in self.wrappers:
            wrapper.on_build_gen(self.cxx_gen_dir)



from rpytest import ft
import pytest

#
# custom_type_caster.h
#


def test_custom_type_caster():
    assert ft.checkConvertRpyintToInt() == 6
    assert ft.convertRpyintToInt(5) == 5
    assert ft.convertRpyintToInt() == 6


#
# type_caster.h
#


def test_get123():
    assert ft.get123() == [1, 2, 3]


#
# type_caster_nested.h
#


def test_nested_typecaster():
    nt = ft.NestedTypecaster()
    ll = None

    def fn(l):
        nonlocal ll
        ll = l

    nt.callWithList(fn)
    assert ll == [1, 2]



from distutils.util import get_platform as _get_platform
from dataclasses import dataclass, field
from typing import List
import re
import typing


# wpilib platforms at https://github.com/wpilibsuite/native-utils/blob/master/src/main/java/edu/wpi/first/nativeutils/WPINativeUtilsExtension.java
@dataclass
class WPILibMavenPlatform:
    arch: str
    os: str = "linux"
    libprefix: str = "lib"

    #: runtime linkage
    libext: str = ".so"

    #: compile time linkage
    linkext: str = None

    #: static linkage
    staticext: str = ".a"

    defines: List[str] = field(default_factory=list)

    def __post_init__(self):
        # linkext defaults to libext
        if self.linkext is None:
            self.linkext = self.libext
        self.defines = [f"{d} 1" for d in self.defines]


X86_64 = "x86-64"

# key is python platform, value is information about wpilib maven artifacts
_platforms = {
    "linux-roborio": WPILibMavenPlatform("athena", defines=["__FRC_ROBORIO__"]),
    "linux-raspbian": WPILibMavenPlatform("arm32", defines=["__RASPBIAN__"]),
    "linux-armv7l": WPILibMavenPlatform("arm32"),
    "linux-x86_64": WPILibMavenPlatform(X86_64),
    "linux-aarch64": WPILibMavenPlatform("arm64"),
    "win32": WPILibMavenPlatform("x86", "windows", "", ".dll", ".lib", ".lib"),
    "win-amd64": WPILibMavenPlatform(X86_64, "windows", "", ".dll", ".lib", ".lib"),
    "macos-universal": WPILibMavenPlatform("universal", "osx", libext=".dylib"),
}


def get_platform_names() -> typing.List[str]:
    return list(_platforms.keys())


def get_platform(name: typing.Optional[str] = None) -> WPILibMavenPlatform:
    """
    Retrieve platform specific information
    """

    # TODO: _PYTHON_HOST_PLATFORM is used for cross builds,
    #       and is returned directly from get_platform. Might
    #       be useful to note for the future.

    if not name:
        pyplatform = _get_platform()

        # Check for 64 bit x86 macOS (version agnostic)
        # - See https://github.com/pypa/setuptools/issues/2520 for universal2
        #   related questions? Sorta.
        if (
            re.fullmatch(r"macosx-.*-x86_64", pyplatform)
            or re.fullmatch(r"macosx-.*-arm64", pyplatform)
            or re.fullmatch(r"macosx-.*-universal2", pyplatform)
        ):
            return _platforms["macos-universal"]

        if pyplatform == "linux-armv7l":
            try:
                import distro

                distro_id = distro.id()

                if distro_id == "raspbian":
                    pyplatform = "linux-raspbian"

            except Exception:
                pass

        elif pyplatform == "linux-armv6":
            try:
                import distro

                distro_id = distro.id()

                if distro_id == "raspbian":
                    pyplatform = "linux-raspbian"
            except Exception:
                pass

        name = pyplatform

    try:
        return _platforms[name]
    except KeyError:
        raise KeyError(f"platform {name} is not supported by robotpy-build!")


def get_platform_override_keys(platform: WPILibMavenPlatform):
    # Used in places where overrides exist
    return [
        f"arch_{platform.arch}",
        f"os_{platform.os}",
        f"platform_{platform.os}_{platform.arch}",
    ]



#
# Portions copied from pybind11's setup_helpers.py
#

import os
from os.path import join
from setuptools.command.build_ext import build_ext
import platform
import setuptools
import sys
import sysconfig
import tempfile

from .util import get_install_root
from ..platforms import get_platform

# TODO: only works for GCC
debug = os.environ.get("RPYBUILD_DEBUG") == "1"

WIN = sys.platform.startswith("win32") and "mingw" not in sysconfig.get_platform()
MACOS = sys.platform.startswith("darwin")
STD_TMPL = "/std:c++{}" if WIN else "-std=c++{}"


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = join(tmpdir, "test.cpp")
        with open(fname, "w") as fp:
            fp.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([fname], output_dir=tmpdir, extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cxx_std(compiler) -> int:
    """Return the -std=c++[11/14/17/20] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """

    for level in (20, 17, 17, 11):
        if has_flag(compiler, STD_TMPL.format(level)):
            return level

    raise RuntimeError("Unsupported compiler -- at least C++11 support is needed!")


def get_opts(typ, std):
    c_opts = {"msvc": ["/EHsc", "/bigobj"], "unix": []}
    l_opts = {"msvc": [], "unix": []}

    plat = get_platform()
    if plat.os == "osx":
        darwin_opts = ["-stdlib=libc++"]

        if "MACOSX_DEPLOYMENT_TARGET" not in os.environ:
            # C++17 requires a higher min version of macOS. An earlier version
            # (10.12 or 10.13) can be set manually via environment variable if
            # you are careful in your feature usage, but 10.14 is the safest
            # setting for general use. However, never set higher than the
            # current macOS version!
            current_macos = tuple(int(x) for x in platform.mac_ver()[0].split(".")[:2])
            if std == 20:
                desired_macos = (10, 15)
            elif std == 17:
                desired_macos = (10, 14)
            else:
                desired_macos = (10, 9)

            macos_string = ".".join(str(x) for x in min(current_macos, desired_macos))
            darwin_opts.append(f"-mmacosx-version-min={macos_string}")

        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts + ["-headerpad_max_install_names"]

    return c_opts.get(typ, []), l_opts.get(typ, [])


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    def build_extensions(self):
        ct = self.compiler.compiler_type
        std = cxx_std(self.compiler)
        opts, link_opts = get_opts(ct, std)

        # To support ccache on windows
        cc_launcher = os.environ.get("RPYBUILD_CC_LAUNCHER")

        if ct == "unix":
            opts.append("-s")  # strip
            if debug:
                opts.append("-ggdb3")
                opts.append("-UNDEBUG")
            else:
                opts.append("-g0")  # remove debug symbols
            opts.append(STD_TMPL.format(std))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")

            if cc_launcher:
                self.compiler.compiler.insert(0, cc_launcher)
                self.compiler.compiler_so.insert(0, cc_launcher)
                # compiler_cxx is only used for linking, so we don't mess with it
                # .. distutils is so weird
                # self.compiler.compiler_cxx.insert(0, cc_launcher)
        elif ct == "msvc":
            opts.append(STD_TMPL.format(std))
            opts.append("/Zc:__cplusplus")
            # Enable standards-compliant preprocessor
            if has_flag(self.compiler, "/Zc:preprocessor"):
                opts.append("/Zc:preprocessor")
            # Enable utf-8 source files.. this probably shouldn't be set globally
            if has_flag(self.compiler, "/utf-8"):
                opts.append("/utf-8")
            if cc_launcher:
                # yes, this is terrible. There's really no other way with distutils
                def _spawn(cmd):
                    if cmd[0] == self.compiler.cc:
                        cmd.insert(0, cc_launcher)
                    self.compiler._rpy_spawn(cmd)

                self.compiler._rpy_spawn = self.compiler.spawn
                self.compiler.spawn = _spawn
        for ext in self.extensions:
            ext.define_macros.append(("PYBIND11_USE_SMART_HOLDER_AS_DEFAULT", "1"))
            if debug:
                ext.define_macros.append(
                    ("PYBIND11_ASSERT_GIL_HELD_INCREF_DECREF", "1")
                )
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts

        # self._gather_global_includes()

        build_ext.build_extensions(self)

        # Fix Libraries on macOS
        # Uses @loader_path, is compatible with macOS >= 10.4
        platform = get_platform()
        if platform.os == "osx":
            from ..relink_libs import relink_extension

            install_root = get_install_root(self)

            for ext in self.extensions:
                libs = relink_extension(
                    install_root,
                    self.get_ext_fullpath(ext.name),
                    self.get_ext_filename(ext.name),
                    ext.rpybuild_wrapper,
                    self.rpybuild_pkgcfg,
                )

                # Used in build_pyi
                ext.rpybuild_libs = libs

    def resolve_libs(self):
        # used in _built_env
        platform = get_platform()
        if platform.os == "osx":
            for wrapper in self.wrappers:
                wrapper.finalize_extension()

            from ..relink_libs import resolve_libs

            install_root = get_install_root(self)

            for ext in self.extensions:
                libs = resolve_libs(
                    install_root,
                    ext.rpybuild_wrapper,
                    self.rpybuild_pkgcfg,
                )

                # Used in build_pyi
                ext.rpybuild_libs = libs

    def run(self):
        # files need to be generated before building can occur
        self.run_command("build_gen")

        for wrapper in self.wrappers:
            wrapper.finalize_extension()

        build_ext.run(self)

        # pyi can only be built after ext is built
        self.run_command("build_pyi")

    def get_libraries(self, ext):
        libraries = build_ext.get_libraries(self, ext)

        if (
            sys.platform != "win32"
            and os.environ.get("RPYBUILD_STRIP_LIBPYTHON") == "1"
        ):
            pythonlib = "python{}.{}".format(
                sys.hexversion >> 24, (sys.hexversion >> 16) & 0xFF
            )
            libraries = [lib for lib in libraries if not lib.startswith(pythonlib)]

        return libraries


parallel = int(os.environ.get("RPYBUILD_PARALLEL", "0"))
if parallel > 0:
    # don't enable this hack by default, because not really sure of the
    # ramifications -- however, it's really useful for development
    #
    # .. the real answer to this is cmake o_O

    # monkey-patch for parallel compilation
    # -> https://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils/13176803#13176803
    def parallelCCompile(
        self,
        sources,
        output_dir=None,
        macros=None,
        include_dirs=None,
        debug=0,
        extra_preargs=None,
        extra_postargs=None,
        depends=None,
    ):
        # those lines are copied from distutils.ccompiler.CCompiler directly
        macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
            output_dir, macros, include_dirs, sources, depends, extra_postargs
        )
        cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
        # parallel code
        import multiprocessing
        import multiprocessing.pool

        N = multiprocessing.cpu_count() if parallel == 1 else parallel

        def _single_compile(obj):
            try:
                src, ext = build[obj]
            except KeyError:
                return
            self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

        for _ in multiprocessing.pool.ThreadPool(N).imap(_single_compile, objects):
            pass
        return objects

    import distutils.ccompiler

    distutils.ccompiler.CCompiler.compile = parallelCCompile



import json
import os
from os.path import exists, dirname, join
import subprocess
import sys

import pybind11_stubgen

try:
    from setuptools.errors import BaseError
except ImportError:
    from distutils.errors import DistutilsError as BaseError

from ._built_env import _BuiltEnv, _PackageFinder


class GeneratePyiError(BaseError):
    pass


class BuildPyi(_BuiltEnv):
    base_package: str

    command_name = "build_pyi"
    description = "Generates pyi files from built extensions"

    def run(self):
        # cannot build pyi files when cross-compiling
        if (
            "_PYTHON_HOST_PLATFORM" in os.environ
            or "PYTHON_CROSSENV" in os.environ
            or os.environ.get("RPYBUILD_SKIP_PYI") == "1"
        ):
            return

        # Gather information for needed stubs
        data, env = self.setup_built_env()
        data["stubs"] = []

        # Ensure that the built extension can always be found
        build_ext = self.get_finalized_command("build_ext")
        for ext in build_ext.extensions:
            data["stubs"].append(ext.name)

        # Don't do anything if nothing is needed
        if not data["stubs"]:
            return

        data_json = json.dumps(data)

        # Execute in a subprocess in case it crashes
        args = [sys.executable, "-m", __name__]
        try:
            subprocess.run(args, input=data_json.encode("utf-8"), env=env, check=True)
        except subprocess.CalledProcessError:
            raise GeneratePyiError(
                "Failed to generate .pyi file (see above, or set RPYBUILD_SKIP_PYI=1 to ignore) via %s"
                % (args,)
            ) from None

        # Create a py.typed for PEP 561
        with open(join(data["out"], *self.base_package.split("."), "py.typed"), "w"):
            pass


def main():
    cfg = json.load(sys.stdin)

    # Configure custom loader
    _PackageFinder.mapping = cfg["mapping"]
    sys.meta_path.insert(0, _PackageFinder)

    # Generate pyi modules
    out = cfg["out"]
    for stub in cfg["stubs"]:
        sys.argv = [
            "<dummy>",
            "--exit-code",
            "--ignore-invalid-expressions=<.*>",
            "--root-suffix=",
            "-o",
            out,
            stub,
        ]

        pybind11_stubgen.main()


if __name__ == "__main__":
    main()



import inspect
import typing as T

from .buffer import RenderBuffer
from .context import (
    ClassContext,
    Documentation,
    EnumContext,
    FunctionContext,
    GeneratedLambda,
    PropContext,
)


def mkdoc(pre: str, doc: Documentation, post: str) -> str:
    if doc:
        if len(doc) == 1:
            return f"{pre}{doc[0]}{post}"

        sep = "\n  "
        return f"{pre}\n  {sep.join(doc)}{post}"

    return ""


def _gensig(cls_qualname: T.Optional[str], fn: FunctionContext) -> str:
    """
    py::overload_cast fails in some obscure cases, so we compute the signature here
    https://github.com/pybind/pybind11/issues/1153
    """
    if fn.const:
        trailing = " const"
    else:
        trailing = ""
    if fn.ref_qualifiers:
        trailing = f"{trailing} {fn.ref_qualifiers}"

    params = ", ".join(param.full_cpp_type for param in fn.all_params)

    return (
        f"{fn.cpp_return_type} ("
        f"{cls_qualname + '::' if cls_qualname and not fn.is_static_method else ''}*)"
        f"({params}){trailing}"
    )


def _genmethod(
    r: RenderBuffer,
    cls_qualname: T.Optional[str],
    fn: FunctionContext,
    trampoline_qualname: T.Optional[str],
    tmpl: str,
):
    qualname = cls_qualname
    arg_params = fn.filtered_params

    if fn.ifdef:
        r.writeln(f"\n#ifdef {fn.ifdef}")
        r.rel_indent(2)

    if fn.ifndef:
        r.writeln(f"\n#ifndef {fn.ifndef}")
        r.rel_indent(2)

    if fn.operator:
        r.writeln(f".def({fn.cpp_code}")
        arg_params = []
    elif fn.is_constructor:
        if fn.cpp_code:
            r.writeln(f".def(py::init({fn.cpp_code})")
        elif fn.genlambda:
            genlambda = fn.genlambda
            arg_params = genlambda.in_params
            lam_params = [param.decl for param in arg_params]
            # TODO: trampoline
            assert cls_qualname is not None
            call_qual = cls_qualname
            _gen_method_lambda(
                r,
                fn,
                genlambda,
                call_qual,
                tmpl,
                f'.def_static("{fn.py_name}"',
                lam_params,
            )
        elif trampoline_qualname:
            r.writeln(
                f".def(py::init_alias<{', '.join(param.full_cpp_type for param in arg_params)}>()"
            )
        else:
            r.writeln(
                f".def(py::init<{', '.join(param.full_cpp_type for param in arg_params)}>()"
            )
    else:
        if fn.is_static_method:
            fn_def = f'.def_static("{fn.py_name}"'
        else:
            fn_def = f'.def("{fn.py_name}"'

        if fn.cpp_code:
            cpp_code = inspect.cleandoc(fn.cpp_code)
            r.writeln(f"{fn_def}, {cpp_code}")

        elif fn.genlambda:
            genlambda = fn.genlambda
            arg_params = genlambda.in_params

            lam_params = [param.decl for param in genlambda.in_params]
            if cls_qualname:
                lam_params = [f"{cls_qualname} &self"] + lam_params

            if trampoline_qualname:
                call_qual = f"(({trampoline_qualname}*)&self)->{fn.cpp_name}"
            elif cls_qualname:
                call_qual = f"(({cls_qualname}*)&self)->{fn.cpp_name}"
            else:
                call_qual = f"{fn.namespace}::{fn.cpp_name}"

            _gen_method_lambda(r, fn, genlambda, call_qual, tmpl, fn_def, lam_params)

        else:
            if trampoline_qualname:
                fn_ns = trampoline_qualname
            elif cls_qualname:
                fn_ns = cls_qualname
            else:
                fn_ns = fn.namespace

            if trampoline_qualname:
                fn_cast = f"static_cast<{_gensig(cls_qualname, fn)}>("
                paren = ")"
            elif fn.is_overloaded:
                fn_cast = f"static_cast<{_gensig(qualname, fn)}>("
                paren = ")"
            else:
                fn_cast = ""
                paren = ""

            if tmpl:
                fn_name = f"template {fn.cpp_name}"
            else:
                fn_name = fn.cpp_name

            r.writeln(f"{fn_def}, {fn_cast}&{fn_ns}::{fn_name}{tmpl}{paren}")

    if arg_params:
        r.writeln(f"  , {', '.join(param.py_arg for param in arg_params)}")

    other_params = []

    if fn.release_gil:
        other_params.append("release_gil()")

    for nurse, patient in fn.keepalives:
        other_params.append(f"py::keep_alive<{nurse}, {patient}>()")

    if fn.return_value_policy:
        other_params.append(fn.return_value_policy)

    if other_params:
        r.writeln(f"  , {', '.join(other_params)}")

    if fn.doc:
        r.writeln(mkdoc("  , py::doc(", fn.doc, ")"))
    r.writeln(")")

    if fn.ifdef:
        r.rel_indent(-2)
        r.writeln(f"#endif // {fn.ifdef}\n")

    if fn.ifndef:
        r.rel_indent(-2)
        r.writeln(f"#endif // {fn.ifndef}\n")


def _gen_method_lambda(
    r: RenderBuffer,
    fn: FunctionContext,
    genlambda: GeneratedLambda,
    call_qual: str,
    tmpl: str,
    fn_def: str,
    lam_params: T.List[str],
):
    r.writeln(f"{fn_def}, []({', '.join(lam_params)}) {{")

    with r.indent():
        if genlambda.pre:
            r.writeln(genlambda.pre)

        call_params = ", ".join(p.call_name for p in fn.all_params)

        r.writeln(f"{genlambda.call_start}{call_qual}{tmpl}({call_params});")

        if genlambda.ret:
            r.writeln(genlambda.ret)

    r.writeln("}")


def genmethod(
    r: RenderBuffer,
    cls_qualname: T.Optional[str],
    fn: FunctionContext,
    trampoline_qualname: T.Optional[str],
):
    if not fn.template_impls:
        _genmethod(r, cls_qualname, fn, trampoline_qualname, "")
    else:
        for tmpl in fn.template_impls:
            _genmethod(
                r,
                cls_qualname,
                fn,
                trampoline_qualname,
                f"<{', '.join(tmpl)}>",
            )


def _genprop(r: RenderBuffer, qualname: str, prop: PropContext):
    doc = ""
    if prop.doc:
        doc = mkdoc(", py::doc(", prop.doc, ")")

    if prop.array_size:
        r.writeln(
            f'.def_property_readonly("{prop.py_name}", []({qualname}& self) {{\n'
            f"   return py::memoryview::from_buffer(\n"
            f"      &self.{prop.cpp_name}, sizeof({prop.cpp_type}),\n"
            f"      py::format_descriptor<{prop.cpp_type}>::value,\n"
            f"      {{{prop.array_size}}}, {{sizeof({prop.cpp_type})}},\n"
            f'      { "true" if prop.readonly else "false" }\n'
            "   );\n"
            f"}}{doc})"
        )
    elif prop.array:
        # cannot sensibly autowrap an array of incomplete size
        pass
    elif prop.reference or prop.bitfield:
        propdef = ".def_property_readonly" if prop.readonly else ".def_property"

        r.writeln(f'{propdef}("{prop.py_name}",')
        with r.indent(2):
            lines = [
                f"[](const {qualname}& self) -> {prop.cpp_type} {{ return self.{prop.cpp_name}; }}"
            ]
            if not prop.readonly:
                lines.append(
                    f"[]({qualname}& self, {prop.cpp_type} v) {{ self.{prop.cpp_name} = v; }}"
                )
            if doc:
                lines.append(doc[2:])

            r.writeln(",\n".join(lines))
        r.writeln(")")

    else:
        propdef = ".def_readonly" if prop.readonly else ".def_readwrite"
        if prop.static:
            propdef = f"{propdef}_static"

        r.writeln(f'{propdef}("{prop.py_name}", &{qualname}::{prop.cpp_name}{doc})')


def enum_decl(r: RenderBuffer, enum: EnumContext, varname: str):
    r.writeln(f"py::enum_<{ enum.full_cpp_name }> {varname};")


def enum_init_args(scope: str, enum: EnumContext):
    params = [scope, f'"{enum.py_name}"']

    if enum.arithmetic:
        params.append("py::arithmetic()")

    if enum.doc:
        params.append(mkdoc("", enum.doc, ""))

    return ", ".join(params)


def enum_def(r: RenderBuffer, varname: str, enum: EnumContext):
    for val in enum.values:
        doc = mkdoc(",", val.doc, "")
        r.writeln(f'.value("{val.py_name}", {val.full_cpp_name}{doc})')

    if enum.inline_code:
        r.write_trim(enum.inline_code)
    r.writeln(";")


def cls_user_using(r: RenderBuffer, cls: ClassContext):
    for typealias in cls.user_typealias:
        r.writeln(f"{typealias};")


def cls_auto_using(r: RenderBuffer, cls: ClassContext):
    for ccls in cls.child_classes:
        if not ccls.template:
            r.writeln(
                f"using {ccls.cpp_name} [[maybe_unused]] = typename {ccls.full_cpp_name};"
            )
    for enum in cls.enums:
        if enum.cpp_name:
            r.writeln(
                f"using {enum.cpp_name} [[maybe_unused]] = typename {enum.full_cpp_name};"
            )
    for typealias in cls.auto_typealias:
        r.writeln(f"{typealias};")


def cls_consts(r: RenderBuffer, cls: ClassContext):
    if cls.constants:
        r.writeln()
        for constant in cls.constants:
            r.writeln(f"static constexpr auto {constant[0]} = {constant[1]};")


def cls_decl(r: RenderBuffer, cls: ClassContext):
    if cls.trampoline:
        tctx = cls.trampoline
        # py::trampoline_self_life_support
        r.write_trim(
            f"""
            struct {tctx.var} : {tctx.full_cpp_name}, py::trampoline_self_life_support {{
                using RpyBase = {tctx.full_cpp_name};
                using RpyBase::RpyBase;
            }};

        """
        )
        r.writeln(
            f'static_assert(std::is_abstract<{tctx.var}>::value == false, "{cls.full_cpp_name} " RPYBUILD_BAD_TRAMPOLINE);'
        )

    class_params = [f"typename {cls.full_cpp_name}"]
    if cls.nodelete:
        class_params.append(
            f"std::unique_ptr<typename {cls.full_cpp_name}, py::nodelete>"
        )

    if cls.trampoline:
        class_params.append(cls.trampoline.var)

    if cls.bases:
        bases = ", ".join(base.full_cpp_name_w_templates for base in cls.bases)
        class_params.append(bases)

    r.writeln(f"py::class_<{', '.join(class_params)}> {cls.var_name};")

    if cls.enums:
        r.writeln()
        for index, enum in enumerate(cls.enums, start=1):
            enum_decl(r, enum, f"{cls.var_name}_enum{index}")

    for ccls in cls.child_classes:
        if not ccls.template:
            cls_decl(r, ccls)


def cls_init(r: RenderBuffer, cls: ClassContext, name: str):

    init_params = [cls.scope_var, name]

    if cls.final:
        init_params.append("py::is_final()")

    if cls.force_multiple_inheritance:
        init_params.append("py::multiple_inheritance()")

    r.writeln(f'{cls.var_name}({", ".join(init_params)}),')

    for idx, enum in enumerate(cls.enums, start=1):
        r.writeln(f"{cls.var_name}_enum{idx}({enum_init_args(cls.var_name, enum)}),")

    for ccls in cls.child_classes:
        if not ccls.template:
            cls_init(r, ccls, f'"{ccls.py_name}"')


def cls_def_enum(r: RenderBuffer, cctx: ClassContext, varname: str):
    for idx, enum in enumerate(cctx.enums, start=1):
        r.writeln(f"{cctx.var_name}_enum{idx}")
        with r.indent():
            enum_def(r, cctx.var_name, enum)


def cls_def(r: RenderBuffer, cls: ClassContext, varname: str):
    if cls.vcheck_fns:
        for fn in cls.vcheck_fns:
            assert fn.cpp_code is not None

            r.writeln("{")
            with r.indent():
                r.writeln(f"auto vcheck = {fn.cpp_code.strip()};")

                sig_params = [f"{cls.full_cpp_name}*"]
                sig_params.extend(p.full_cpp_type for p in fn.all_params)

                signature = f"{fn.cpp_return_type}({', '.join(sig_params)})"
                r.writeln(
                    f"static_assert(std::is_convertible<decltype(vcheck), std::function<{signature}>>::value,"
                )
                r.writeln(
                    f'  "{cls.full_cpp_name}::{fn.cpp_name} must have virtual_xform if cpp_code signature doesn\'t match original function");'
                )
            r.writeln("}")

    if cls.doc:
        r.writeln(f'{varname}.doc() = {mkdoc("", cls.doc, "")};')

    r.writeln(varname)
    with r.indent():

        if cls.add_default_constructor:
            r.writeln(".def(py::init<>(), release_gil())")

        for fn in cls.wrapped_public_methods:
            genmethod(r, cls.full_cpp_name, fn, None)

        if cls.trampoline is not None:
            for fn in cls.wrapped_protected_methods:
                genmethod(r, cls.full_cpp_name, fn, cls.trampoline.var)

        for prop in cls.public_properties:
            _genprop(r, cls.full_cpp_name, prop)

        if cls.trampoline is not None:
            for prop in cls.protected_properties:
                _genprop(r, cls.trampoline.full_cpp_name, prop)

    if cls.inline_code:
        r.writeln()
        r.write_trim(cls.inline_code)

    r.writeln(";")

    if cls.unnamed_enums:
        r.writeln()
        for enum in cls.unnamed_enums:
            for val in enum.values:
                r.writeln(
                    f'{varname}.attr("{val.py_name}") = (int){val.full_cpp_name};'
                )

    for ccls in cls.child_classes:
        if not ccls.template:
            cls_def(r, ccls, ccls.var_name)
